#!/usr/bin/env python3
# TextMIDITools Version 1.0.18
# textmidiform.py 1.0
# Copyright © 2022 Thomas E. Janzen
# License GPLv3+: GNU GPL version 3 or later <https://gnu.org/licenses/gpl.html>
# This is free software: you are free to change and redistribute it.
# There is NO WARRANTY, to the extent permitted by law.
#

from Sine import Sine

import tkinter
import tkinter.ttk
from tkinter import *
from tkinter.ttk import *

class FormFrame(tkinter.Frame):
    def __init__(self, parent, form_title, mean_period_callback, mean_phase_callback, range_period_callback, range_phase_callback, xml_subform, mean_on = True, range_on = True):
        self.the_row = 0;
        self.mean_period_callback  = mean_period_callback
        self.mean_phase_callback   = mean_phase_callback
        self.range_period_callback = range_period_callback
        self.range_phase_callback  = range_phase_callback
        self.xml_subform = xml_subform
        self.mean_on = mean_on
        self.range_on = range_on
        super().__init__(parent)
        self.grid(sticky=NSEW, padx=10, pady=10)
        self.form_title = form_title
        self.create_widgets()

    def create_widgets(self):
        self.form_title_label = tkinter.ttk.Label(self, text=self.form_title)
        self.form_title_label.grid(row=0, column=0)

        if (self.mean_on):
            self.mean = Sine(parent=self, sine_title="Mean", period_callback=self.mean_period_callback, phase_callback=self.mean_phase_callback, xml_sine=self.xml_subform['mean'])
            self.mean.grid(row=self.the_row, column=2, sticky=NSEW, padx=10, pady=10)
            self.the_row =  self.the_row + 1
        if (self.range_on):
            self.range = Sine(parent=self, sine_title="Range", period_callback=self.range_period_callback, phase_callback=self.range_phase_callback, xml_sine=self.xml_subform['range'])
            self.range.grid(row=self.the_row, column=2, sticky=NSEW, padx=10, pady=10)

    def install_xml_subform(self, xml_subform):
        self.xml_subform = xml_subform
        self.mean.install_xml_sine(xml_subform['mean'])
        self.range.install_xml_sine(xml_subform['range'])
        self.update()




#!/usr/bin/env python3
# TextMIDITools Version 1.0.18
# Copyright © 2022 Thomas E. Janzen
# License GPLv3+: GNU GPL version 3 or later <https://gnu.org/licenses/gpl.html>
# This is free software: you are free to change and redistribute it.
# There is NO WARRANTY, to the extent permitted by law.
#

import re
import math
import sys
import tkinter
import tkinter.ttk
from tkinter import *
from tkinter import ttk
from tkinter.ttk import *
import tkinter.constants, tkinter.filedialog
from xml.dom import minidom
from xml.dom.minidom import parse, Node, getDOMImplementation
from VoiceWindow import VoiceWindow
from AllFormsWindow import AllFormsWindow
from AllFormsWindow import ScaleFrame

class XmlForm(tkinter.Tk):
    xml_form_dict = {}
    twopi = 2.0 * math.pi
    the_filename = "Untitled"
    def __init__(self):
        super().__init__()
        self.default_xml_form()
        self.all_forms_window = AllFormsWindow(self.xml_form_dict)
        self.voice_window = VoiceWindow(self.xml_form_dict)
        self.win_height = 600
        self.win_width = 1000
        self.canvas = Canvas(self, bg='white', height=self.win_height, width=self.win_width, background='#88AAFF')
        self.canvas.grid(row=0, column=0, sticky=(N,S,E,W))
        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)

        self.canvas.bind("<Configure>", self.configure_callback)

        self.draw_form()
        self.title('Form Plot')

        self.frame = ttk.Frame(self, padding="1 1 1 1")

        self.option_add('*tearOff', FALSE)
        self.frame.grid(column=0, row=0, sticky=(N, W, E, S))
        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)
        self.frame.focus()
        self.makemenu(self)

        self.frame.grid(sticky="we", row=0, column=2)
        self.frame.rowconfigure(index=0, weight=1)
        self.frame.columnconfigure(index=2, weight=1)
        self.geometry('1000x600+435+50')
        #self.frame.grab_set()

    def install_file(self, afilename):
        self.the_filename = afilename
        # get the xml.dom.minidom.Document
        self.dom = parse(self.the_filename)
        self.traverse_xml_form()
        self.draw_form()

    def configure_callback(self, event=None):
        self.win_width = event.width
        self.win_height = event.height
        self.draw_form()

    def draw_form(self):
        duration = float(self.xml_form_dict['len'])

        self.canvas.delete('all')
        self.canvas.grid(row=0, column=0, sticky=NSEW)

        self.canvas.grid(sticky=NSEW)

        poly_lr = []
        poly_rl = []
        for i in range(1, self.win_width, 1):
            mean_freq  = 1.0 / float(self.xml_form_dict['pitch_form']['mean']['period'])
            range_freq = 1.0 / float(self.xml_form_dict['pitch_form']['range']['period'])
            seconds_per_pixel = duration / float(self.win_width)
            the_time = float(i) * seconds_per_pixel
            # omega * t = twopi * mean_freq  * the_time
            # the phase = self.xml_form_dict['pitch_form']['mean']['phase']
            # omega * t + phase = twopi * mean_freq  * the_time + self.xml_form_dict['pitch_form']['mean']['phase']
            # take sin(omega*t + phase) which ranges -1.0 to 1.0
            # make it work with the form model by making it range from 0 to 1.
            # 0.5 + sin(omega * t + phase) / 2
            # scale it for graphical plotting on the canvas
            # 0.5 + sin(omega * t + phase) / 2 * self.win_height / 16
            # but window coordinates are left-handed, so invert the mean curve
            # -(0.5 + sin(omega * t + phase) / 2 * self.win_height / 16)
            # position it on the canvas
            # 0.5 + sin(omega * t + phase) / 2 * self.win_height / 16 + 1 * self.win_height / 16
            # oi la!
            y_mean  = (.5 - math.sin(self.twopi * mean_freq  * the_time + float(self.xml_form_dict['pitch_form']['mean']['phase']))  / 2.0) * (float(self.win_height) / 12.0) + (1.0 * float(self.win_height) / 16.0)
            y_range = (.5 + math.sin(self.twopi * range_freq * the_time + float(self.xml_form_dict['pitch_form']['range']['phase'])) / 2.0) * (float(self.win_height) / 12.0)

            poly_lr.append(i)
            poly_lr.append(y_mean - y_range / 2)
            poly_rl.append(y_mean + y_range / 2)
            poly_rl.append(i)
        poly_rl.reverse()
        polygon = poly_lr + poly_rl
        fill_color = '#004400'
        pitch_plot_id = self.canvas.create_polygon(polygon, fill=fill_color)
        y_zero = (.5 - 0  / 2.0) * (float(self.win_height) / 12.0) + (1.0 * float(self.win_height) / 16.0)
        axis_color = '#000000'
        self.canvas.create_line(0, y_zero, self.win_width, y_zero, fill=axis_color, dash='.')

        poly_lr = []
        poly_rl = []
        polygon = []
        for i in range(1, self.win_width):
            mean_freq = 1.0 / float(self.xml_form_dict['rhythm_form']['mean']['period'])
            range_freq = 1.0 / float(self.xml_form_dict['rhythm_form']['range']['period'])
            seconds_per_pixel = duration / float(self.win_width)
            the_time = float(i) * seconds_per_pixel
            y_mean  = (.5
                       - math.sin(self.twopi * mean_freq  * the_time
                                  + float(self.xml_form_dict['rhythm_form']['mean']['phase']))
                       / 2.0) * float(self.win_height) / 12.0 + 5.0 * float(self.win_height / 16.0)
            y_range = (.5
                      + math.sin(self.twopi * range_freq * the_time
                                 + float(self.xml_form_dict['rhythm_form']['range']['phase']))
                      / 2.0) * float(self.win_height) / 12.0
            poly_lr.append(i)
            poly_lr.append(y_mean - y_range / 2)
            poly_rl.append(y_mean + y_range / 2)
            poly_rl.append(i)
        poly_rl.reverse()
        polygon = poly_lr + poly_rl
        rhythm_plot_id = self.canvas.create_polygon(polygon, fill=fill_color)
        y_zero = (.5 - 0  / 2.0) * (float(self.win_height) / 12.0) + (5.0 * float(self.win_height) / 16.0)
        self.canvas.create_line(0, y_zero, self.win_width, y_zero, fill=axis_color, dash='.')

        poly_lr = []
        poly_rl = []
        polygon = []
        for i in range(1, self.win_width):
            mean_freq = 1.0 / float(self.xml_form_dict['dynamic_form']['mean']['period'])
            range_freq = 1.0 / float(self.xml_form_dict['dynamic_form']['range']['period'])
            seconds_per_pixel = duration / float(self.win_width)
            the_time = float(i) * seconds_per_pixel
            y_mean  = (.5
                       - math.sin(self.twopi * mean_freq  * the_time
                                  + float(self.xml_form_dict['dynamic_form']['mean']['phase']))
                       / 2.0) * float(self.win_height) / 12.0 + 9.0 * float(self.win_height / 16.0)
            y_range = (.5
                      + math.sin(self.twopi * range_freq * the_time
                                 + float(self.xml_form_dict['dynamic_form']['range']['phase']))
                      / 2.0) * float(self.win_height) / 12.0
            poly_lr.append(i)
            poly_lr.append(y_mean - y_range / 2)
            poly_rl.append(y_mean + y_range / 2)
            poly_rl.append(i)
        poly_rl.reverse()
        polygon = poly_lr + poly_rl
        dynamic_plot_id = self.canvas.create_polygon(polygon, fill=fill_color)
        y_zero = (.5 - 0  / 2.0) * (float(self.win_height) / 12.0) + (9.0 * float(self.win_height) / 16.0)
        self.canvas.create_line(0, y_zero, self.win_width, y_zero, fill=axis_color, dash='.')

        poly_lr = []
        poly_rl = []
        polygon = []
        # Texture (number of voices playing) was misdefined as the mean alone,
        # but really it is the range.  So there is a reversal of roles here.
        for i in range(1, self.win_width):
            mean_freq = 1.0 / float(self.xml_form_dict['texture_form']['period'])
            seconds_per_pixel = duration / float(self.win_width)
            the_time = float(i) * seconds_per_pixel
            y_mean  = (.5
                      + math.sin(self.twopi * mean_freq  * the_time
                                 + float(self.xml_form_dict['texture_form']['phase']))
                      / 2.0) * float(self.win_height / 12.0)
            y_range = float(self.win_height) / 12.0 + (13.0 * float(self.win_height) / 16.0)
            poly_lr.append(i)
            poly_lr.append(y_range - y_mean / 2)
            poly_rl.append(y_range + y_mean / 2)
            poly_rl.append(i)
        poly_rl.reverse()
        polygon = poly_lr + poly_rl
        texture_plot_id = self.canvas.create_polygon(polygon, fill=fill_color)
        y_zero = (.5 - 0  / 2.0) * (float(self.win_height) / 12.0) + (float(self.win_height) / 16.0)
        self.canvas.create_line(0, y_zero, self.win_width, y_zero, fill=axis_color, dash='.')

        # Draw minute markers
        scale_height = self.win_height * 55 / 100
        #self.canvas.create_line(0, scale_height, self.win_width, scale_height)
        for s in range(0, int(duration), 60):
            tickx = s / seconds_per_pixel
            self.canvas.create_line(tickx, 0, tickx, self.win_height, fill=axis_color, dash='.')

        widget = tkinter.ttk.Label(self.canvas, text='Pitch', style="BW.TLabel")

        self.canvas.create_window(self.win_width, self.win_width, window=widget)
        self.canvas.create_text(self.win_width / 2, 20, text=self.xml_form_dict["name"], fill=axis_color)
        self.canvas.create_text(30, self.win_height * 1 / 16 - 30, text='Pitch', fill=axis_color)
        self.canvas.create_text(30, self.win_height * 5 / 16 - 30, text='Rhythm', fill=axis_color)
        self.canvas.create_text(30, self.win_height * 9 / 16  - 30, text='Dynamic', fill=axis_color)
        self.canvas.create_text(30, self.win_height * 13 / 16 - 30, text='Texture', fill=axis_color)

    def default_melody_probabilities(self):
        # These probabilities are cumulative.
        # 1.0 - up == same - down,  so that the walking voice
        #            goes up or down equal amounts.
        #     # up
        #     # up
        # .75 - up if above up
        #     # same
        #     # same
        # .50 - same if above same
        #     # down
        #     # down
        # .25 - down if above down
        #     # (silent)
        #     # (silent)
        melody_probabilities_dict = {}
        melody_probabilities_dict['down'] = 0.25
        melody_probabilities_dict['same'] = 0.5
        melody_probabilities_dict['up'] = 0.75
        return melody_probabilities_dict

    def default_form(self):
        pd = {}
        pmd = {}
        pmd['period'] = 180.0
        pmd['phase']  = 0.0
        pmd['amplitude'] = 0.50
        pmd['offset'] = 0.50
        pd['mean'] = pmd

        prd = {}
        prd['period'] = 180.0
        prd['phase'] = 0.0
        prd['amplitude'] = 0.50
        prd['offset'] = 0.50
        pd['range'] = prd
        return pd

    def default_texture_form(self):
        td = {}
        td['period'] = 180.0
        td['phase'] = 0.0
        td['amplitude'] = 0.50
        td['offset'] = 0.50
        return td

    def default_voices(self):
        vl = []
        for v in range(1, 24):
            vd = {}
            vd['low_pitch'] = "A0"
            vd['high_pitch'] = "C8"
            vd['channel'] = 1
            vd['walking'] = 0
            vd['program'] = 1
            vd['pan'] = 0
            fd = {}
            fd['follow'] = 0
            fd['leader'] = 2147483647
            fd['interval_type'] = 0
            fd['interval'] = 0
            vd['follower'] = fd
            vl.append(vd)

        return vl

    def default_xml_form(self):
        self.xml_form_dict['name'] = "Wave Sonata"
        self.xml_form_dict['len'] = 600.0
        self.xml_form_dict['min_note_len'] = 0.0
        self.xml_form_dict['max_note_len'] = 2.0
        self.xml_form_dict['scale'] = ScaleFrame.full_midi_scale
        self.xml_form_dict['pulse'] = 8.0
        self.xml_form_dict['melody_probabilities'] = self.default_melody_probabilities()
        self.xml_form_dict['pitch_form'] = self.default_form()
        self.xml_form_dict['rhythm_form'] = self.default_form()
        self.xml_form_dict['dynamic_form'] = self.default_form()
        self.xml_form_dict['texture_form'] = self.default_texture_form()
        self.xml_form_dict['voices'] = self.default_voices()

    def traverse_scale(self, scale):
        scale_array = []
        scale_list = scale.getElementsByTagName('item')

        for item_node in scale_list :
            scale_array.append(item_node.firstChild.data)
        return scale_array

    def traverse_melody_probabilities(self, melody_probabilities):
        melody_probabilities_dict = {}
        down_element_list = melody_probabilities.getElementsByTagName('down_')
        melody_probabilities_dict['down'] = down_element_list[0].firstChild.data
        same_element_list = melody_probabilities.getElementsByTagName('same_')
        melody_probabilities_dict['same'] = same_element_list[0].firstChild.data
        up_element_list = melody_probabilities.getElementsByTagName('up_')
        melody_probabilities_dict['up'] = up_element_list[0].firstChild.data
        return melody_probabilities_dict

    def traverse_pitch_form(self, pitch_form):
        pd = {}
        pmd = {}
        mean_sine_node = pitch_form.getElementsByTagName("mean_sine_")[0]
        mean_period_node    = mean_sine_node.getElementsByTagName("period_")[0]
        mean_period = mean_period_node.firstChild.data
        pmd['period'] = float(mean_period)
        mean_phase_node     = mean_sine_node.getElementsByTagName("phase_")[0]
        mean_phase = mean_phase_node.firstChild.data
        pmd['phase'] = float(mean_phase)
        mean_amplitude_node = mean_sine_node.getElementsByTagName("amplitude_")[0]
        mean_amplitude = mean_amplitude_node.firstChild.data
        pmd['amplitude'] = float(mean_amplitude)
        mean_offset_node    = mean_sine_node.getElementsByTagName("offset_")[0]
        mean_offset = mean_offset_node.firstChild.data
        pmd['offset'] = float(mean_offset)
        pd['mean'] = pmd

        prd = {}
        range_sine_node = pitch_form.getElementsByTagName("range_sine_")[0]
        range_period_node    = range_sine_node.getElementsByTagName("period_")[0]
        range_period = range_period_node.firstChild.data
        prd['period'] = float(range_period)
        range_phase_node     = range_sine_node.getElementsByTagName("phase_")[0]
        range_phase = range_phase_node.firstChild.data
        prd['phase'] = float(range_phase)
        range_amplitude_node = range_sine_node.getElementsByTagName("amplitude_")[0]
        range_amplitude = range_amplitude_node.firstChild.data
        prd['amplitude'] = float(range_amplitude)
        range_offset_node    = range_sine_node.getElementsByTagName("offset_")[0]
        range_offset = range_offset_node.firstChild.data
        prd['offset'] = float(range_offset)
        pd['range'] = prd
        return pd

    def traverse_rhythm_form(self, rhythm_form):
        rd = {}
        rmd = {}
        mean_sine_node = rhythm_form.getElementsByTagName("mean_sine_")[0]
        mean_period_node    = mean_sine_node.getElementsByTagName("period_")[0]
        mean_period = mean_period_node.firstChild.data
        rmd['period'] = float(mean_period)
        mean_phase_node     = mean_sine_node.getElementsByTagName("phase_")[0]
        mean_phase = mean_phase_node.firstChild.data
        rmd['phase'] = float(mean_phase)
        mean_amplitude_node = mean_sine_node.getElementsByTagName("amplitude_")[0]
        mean_amplitude = mean_amplitude_node.firstChild.data
        rmd['amplitude'] = float(mean_amplitude)
        mean_offset_node    = mean_sine_node.getElementsByTagName("offset_")[0]
        mean_offset = mean_offset_node.firstChild.data
        rmd['offset'] = float(mean_offset)
        rd['mean'] = rmd

        rrd = {}
        range_sine_node = rhythm_form.getElementsByTagName("range_sine_")[0]
        range_period_node    = range_sine_node.getElementsByTagName("period_")[0]
        range_period = range_period_node.firstChild.data
        rrd['period'] = float(range_period)
        range_phase_node     = range_sine_node.getElementsByTagName("phase_")[0]
        range_phase = range_phase_node.firstChild.data
        rrd['phase'] = float(range_phase)
        range_amplitude_node = range_sine_node.getElementsByTagName("amplitude_")[0]
        range_amplitude = range_amplitude_node.firstChild.data
        rrd['amplitude'] = float(range_amplitude)
        range_offset_node    = range_sine_node.getElementsByTagName("offset_")[0]
        range_offset = range_offset_node.firstChild.data
        rrd['offset'] = float(range_offset)
        rd['range'] = rrd
        return rd

    def traverse_dynamic_form(self, dynamic_form):
        dd = {}
        dmd = {}
        mean_sine_node = dynamic_form.getElementsByTagName("mean_sine_")[0]
        mean_period_node    = mean_sine_node.getElementsByTagName("period_")[0]
        mean_period = mean_period_node.firstChild.data
        dmd['period'] = float(mean_period)
        mean_phase_node     = mean_sine_node.getElementsByTagName("phase_")[0]
        mean_phase = mean_phase_node.firstChild.data
        dmd['phase'] = float(mean_phase)
        mean_amplitude_node = mean_sine_node.getElementsByTagName("amplitude_")[0]
        mean_amplitude = mean_amplitude_node.firstChild.data
        dmd['amplitude'] = float(mean_amplitude)
        mean_offset_node    = mean_sine_node.getElementsByTagName("offset_")[0]
        mean_offset = mean_offset_node.firstChild.data
        dmd['offset'] = float(mean_offset)
        dd['mean'] = dmd

        drd = {}
        range_sine_node = dynamic_form.getElementsByTagName("range_sine_")[0]
        range_period_node    = range_sine_node.getElementsByTagName("period_")[0]
        range_period = range_period_node.firstChild.data
        drd['period'] = float(range_period)
        range_phase_node     = range_sine_node.getElementsByTagName("phase_")[0]
        range_phase = range_phase_node.firstChild.data
        drd['phase'] = float(range_phase)
        range_amplitude_node = range_sine_node.getElementsByTagName("amplitude_")[0]
        range_amplitude = range_amplitude_node.firstChild.data
        drd['amplitude'] = float(range_amplitude)
        range_offset_node    = range_sine_node.getElementsByTagName("offset_")[0]
        range_offset = range_offset_node.firstChild.data
        drd['offset'] = float(range_offset)
        dd['range'] = drd
        return dd

    def traverse_texture_form(self, texture_form):
        td = {}
        period_node    = texture_form.getElementsByTagName("period_")[0]
        period = period_node.firstChild.data
        td['period'] = float(period)
        phase_node     = texture_form.getElementsByTagName("phase_")[0]
        phase = phase_node.firstChild.data
        td['phase'] = float(phase)
        amplitude_node = texture_form.getElementsByTagName("amplitude_")[0]
        amplitude = amplitude_node.firstChild.data
        td['amplitude'] = float(amplitude)
        offset_node    = texture_form.getElementsByTagName("offset_")[0]
        offset = offset_node.firstChild.data
        td['offset'] = float(offset)
        return td

    def traverse_voices(self, voices):
        voice_list = voices.getElementsByTagName("item")
        vl = []
        for v in voice_list:
            vd = {}
            low_pitch_node = v.getElementsByTagName("low_pitch_")[0]
            low_pitch = low_pitch_node.firstChild.data
            vd['low_pitch'] = low_pitch
            high_pitch_node = v.getElementsByTagName("high_pitch_")[0]
            high_pitch = high_pitch_node.firstChild.data
            vd['high_pitch'] = high_pitch
            channel_node = v.getElementsByTagName("channel_")[0]
            channel = channel_node.firstChild.data
            vd['channel'] = int(channel)
            walking_node = v.getElementsByTagName("walking_")[0]
            walking = walking_node.firstChild.data
            vd['walking'] = int(walking)
            program_node = v.getElementsByTagName("program_")[0]
            program = program_node.firstChild.data
            vd['program'] = int(program)
            pan_node = v.getElementsByTagName("pan_")[0]
            pan = pan_node.firstChild.data
            vd['pan'] = pan
            fd = {}
            follower_node = v.getElementsByTagName("follower_")[0]
            follow_node = follower_node.getElementsByTagName("follow_")[0]
            follow = follow_node.firstChild.data
            fd['follow'] = int(follow)
            leader_node = follower_node.getElementsByTagName("leader_")[0]
            leader = leader_node.firstChild.data
            fd['leader'] = int(leader)
            interval_type_node = follower_node.getElementsByTagName("interval_type_")[0]
            interval_type = interval_type_node.firstChild.data
            fd['interval_type'] = int(interval_type)
            interval_node = follower_node.getElementsByTagName("interval_")[0]
            interval = interval_node.firstChild.data
            fd['interval'] = int(interval)
            vd['follower'] = fd
            vl.append(vd)

        return vl

    def traverse_xml_form(self):
        self.xml_form_dict['name'] = self.dom.getElementsByTagName("name_")[0].firstChild.data
        self.xml_form_dict['len'] = float(self.dom.getElementsByTagName("len_")[0].firstChild.data)
        self.xml_form_dict['min_note_len'] = self.dom.getElementsByTagName("min_note_len_")[0].firstChild.data
        self.xml_form_dict['max_note_len'] = self.dom.getElementsByTagName("max_note_len_")[0].firstChild.data

        self.xml_form_dict['scale'] = self.traverse_scale(self.dom.getElementsByTagName("scale_")[0])
        self.xml_form_dict['pulse'] = self.dom.getElementsByTagName("pulse_")[0].firstChild.data
        self.xml_form_dict['melody_probabilities'] = self.traverse_melody_probabilities(self.dom.getElementsByTagName("melody_probabilities_")[0])

        self.xml_form_dict['pitch_form'] = self.traverse_pitch_form(self.dom.getElementsByTagName("pitch_form_")[0])
        self.xml_form_dict['rhythm_form'] = self.traverse_rhythm_form(self.dom.getElementsByTagName("rhythm_form_")[0])
        self.xml_form_dict['dynamic_form'] = self.traverse_dynamic_form(self.dom.getElementsByTagName("dynamic_form_")[0])
        self.xml_form_dict['texture_form'] = self.traverse_texture_form(self.dom.getElementsByTagName("texture_form_")[0])
        self.xml_form_dict['voices'] = self.traverse_voices(self.dom.getElementsByTagName("voices_")[0])

    def postscript_callback(self):
        afilename=tkinter.filedialog.asksaveasfilename(initialdir = ".",title = 'Save Postscript File', filetypes=(("ps files","*.ps"),("all files","*.*")))
        self.canvas.postscript(file=afilename, colormode='gray', pageheight='8i', pagewidth='8i')

    def makemenu(self, toplevelwin):
        top_menu = Menu(toplevelwin)
        toplevelwin['menu'] = top_menu;
        file_menu = Menu(top_menu)
        top_menu.add_cascade(label='File', menu=file_menu, underline=0)
        file_menu.add_command(label='Open...',   command=self.get_file_callback, underline=0, accelerator='O')
        file_menu.add_command(label='Save...',   command=self.save_callback, underline=0, accelerator='S')
        file_menu.add_command(label='Redraw', command=self.redraw_callback, underline=0, accelerator='R')
        file_menu.add_command(label='Save Postscript...', command=self.postscript_callback,
            underline=0, accelerator='S')
        file_menu.add_command(label='About...',   command=self.about_callback, underline=0, accelerator='A')
        file_menu.add_command(label='Quit', command=lambda:toplevelwin.quit(), underline=0, accelerator='Q')

    def get_file_callback(self):
        afilename=tkinter.filedialog.askopenfilename(initialdir = ".",title = 'Select XML Form File', filetypes=(("xml files","*.xml"),("all files","*.*")))
        self.install_file(afilename)
        self.voice_window.install_xml_form(self.xml_form_dict)
        self.all_forms_window.install_xml_form(self.xml_form_dict)

    def save_callback(self):
        afilename=tkinter.filedialog.asksaveasfilename(initialdir = ".",title = 'Select XML Form File', filetypes=(("xml files","*.xml"),("all files","*.*")))
        dom_impl = getDOMImplementation()
        # copy form parts of xml_form_dict from AllForms
        # copy voice parts from Voice

        self.xml_form_dict['name'] = self.all_forms_window.xml_form['name']
        self.xml_form_dict['len'] = self.all_forms_window.xml_form['len']
        self.xml_form_dict['min_note_len'] = self.all_forms_window.xml_form['min_note_len']
        self.xml_form_dict['max_note_len'] = self.all_forms_window.xml_form['max_note_len']
        # scale
        self.xml_form_dict['scale'] = self.all_forms_window.xml_form['scale']
        self.xml_form_dict['pulse'] = self.all_forms_window.xml_form['pulse']
        self.xml_form_dict['melody_probabilities']['down'] = self.all_forms_window.xml_form['melody_probabilities']['down']
        self.xml_form_dict['melody_probabilities']['same'] = self.all_forms_window.xml_form['melody_probabilities']['same']
        self.xml_form_dict['melody_probabilities']['up'] = self.all_forms_window.xml_form['melody_probabilities']['up']

        self.xml_form_dict['pitch_form']['mean']['period'] = self.all_forms_window.xml_form['pitch_form']['mean']['period']
        self.xml_form_dict['pitch_form']['mean']['phase'] = self.all_forms_window.xml_form['pitch_form']['mean']['phase']
        self.xml_form_dict['pitch_form']['mean']['amplitude'] = self.all_forms_window.xml_form['pitch_form']['mean']['amplitude']
        self.xml_form_dict['pitch_form']['mean']['offset'] = self.all_forms_window.xml_form['pitch_form']['mean']['offset']

        self.xml_form_dict['rhythm_form']['mean']['period'] = self.all_forms_window.xml_form['rhythm_form']['mean']['period']
        self.xml_form_dict['rhythm_form']['mean']['phase'] = self.all_forms_window.xml_form['rhythm_form']['mean']['phase']
        self.xml_form_dict['rhythm_form']['mean']['amplitude'] = self.all_forms_window.xml_form['rhythm_form']['mean']['amplitude']
        self.xml_form_dict['rhythm_form']['mean']['offset'] = self.all_forms_window.xml_form['rhythm_form']['mean']['offset']

        self.xml_form_dict['dynamic_form']['mean']['period'] = self.all_forms_window.xml_form['dynamic_form']['mean']['period']
        self.xml_form_dict['dynamic_form']['mean']['phase'] = self.all_forms_window.xml_form['dynamic_form']['mean']['phase']
        self.xml_form_dict['dynamic_form']['mean']['amplitude'] = self.all_forms_window.xml_form['dynamic_form']['mean']['amplitude']
        self.xml_form_dict['dynamic_form']['mean']['offset'] = self.all_forms_window.xml_form['dynamic_form']['mean']['offset']

        self.xml_form_dict['texture_form']['period'] = self.all_forms_window.xml_form['texture_form']['period']
        self.xml_form_dict['texture_form']['phase'] = self.all_forms_window.xml_form['texture_form']['phase']
        self.xml_form_dict['texture_form']['amplitude'] = self.all_forms_window.xml_form['texture_form']['amplitude']
        self.xml_form_dict['texture_form']['offset'] = self.all_forms_window.xml_form['texture_form']['offset']

        self.xml_form_dict['pitch_form']['range']['period'] = self.all_forms_window.xml_form['pitch_form']['range']['period']
        self.xml_form_dict['pitch_form']['range']['phase'] = self.all_forms_window.xml_form['pitch_form']['range']['phase']
        self.xml_form_dict['pitch_form']['range']['amplitude'] = self.all_forms_window.xml_form['pitch_form']['range']['amplitude']
        self.xml_form_dict['pitch_form']['range']['offset'] = self.all_forms_window.xml_form['pitch_form']['range']['offset']

        self.xml_form_dict['rhythm_form']['range']['period'] = self.all_forms_window.xml_form['rhythm_form']['range']['period']
        self.xml_form_dict['rhythm_form']['range']['phase'] = self.all_forms_window.xml_form['rhythm_form']['range']['phase']
        self.xml_form_dict['rhythm_form']['range']['amplitude'] = self.all_forms_window.xml_form['rhythm_form']['range']['amplitude']
        self.xml_form_dict['rhythm_form']['range']['offset'] = self.all_forms_window.xml_form['rhythm_form']['range']['offset']

        self.xml_form_dict['dynamic_form']['range']['period'] = self.all_forms_window.xml_form['dynamic_form']['range']['period']
        self.xml_form_dict['dynamic_form']['range']['phase'] = self.all_forms_window.xml_form['dynamic_form']['range']['phase']
        self.xml_form_dict['dynamic_form']['range']['amplitude'] = self.all_forms_window.xml_form['dynamic_form']['range']['amplitude']
        self.xml_form_dict['dynamic_form']['range']['offset'] = self.all_forms_window.xml_form['dynamic_form']['range']['offset']

        form_document_type = dom_impl.createDocumentType("boost_serialization", None, None)
        form_document = dom_impl.createDocument(None, "boost_serialization", form_document_type)
        top = form_document.documentElement
        top.setAttribute("signature", 'serialization::archive')
        top.setAttribute("version", '14')
        xml_form_element = form_document.createElement("xml_form")
        xml_form_element.setAttribute("class_id", '0')
        xml_form_element.setAttribute("tracking_level", '0')
        xml_form_element.setAttribute("version", '0')
        top.appendChild(xml_form_element)

        self.add_text_element(form_document, xml_form_element, "name")
        self.add_text_element(form_document, xml_form_element, "len")
        self.add_text_element(form_document, xml_form_element, "min_note_len")
        self.add_text_element(form_document, xml_form_element, "max_note_len")

        scale_element = form_document.createElement("scale_")
        scale_element.setAttribute("class_id", '1')
        scale_element.setAttribute("tracking_level", '0')
        scale_element.setAttribute("version", '0')

        temp_scale = self.all_forms_window.scale_frame.get_string_scale()
        self.add_text_element(form_document, scale_element, "count", "count", str(len(temp_scale)))

        self.add_text_element(form_document, scale_element, "item_version", "item_version", "0")

        self.xml_form_dict['scale'] = []
        self.xml_form_dict['scale'] = self.all_forms_window.scale_frame.get_string_scale()
        for n in self.xml_form_dict['scale']:
            self.add_text_element(form_document, scale_element, "item", "item", n)

        xml_form_element.appendChild(scale_element)

        self.add_text_element(form_document, xml_form_element, "pulse")

        melody_probabilities_element = form_document.createElement("melody_probabilities_")
        melody_probabilities_element.setAttribute("class_id", '2')
        melody_probabilities_element.setAttribute("tracking_level", '0')
        melody_probabilities_element.setAttribute("version", '0')

        self.add_text_element(form_document, melody_probabilities_element, "down", "down_", str(self.xml_form_dict['melody_probabilities']['down']))
        self.add_text_element(form_document, melody_probabilities_element, "same", "same_", str(self.xml_form_dict['melody_probabilities']['same']))
        self.add_text_element(form_document, melody_probabilities_element, "up", "up_", str(self.xml_form_dict['melody_probabilities']['up']))
        xml_form_element.appendChild(melody_probabilities_element)

        self.add_form_element(form_document, xml_form_element, "pitch_form", True)
        self.add_form_element(form_document, xml_form_element, "rhythm_form")
        self.add_form_element(form_document, xml_form_element, "dynamic_form")
        self.add_sine_element(form_document, xml_form_element, "texture_form")

        voices_element = form_document.createElement("voices_")
        voices_element.setAttribute("class_id", '5')
        voices_element.setAttribute("tracking_level", '0')
        voices_element.setAttribute("version", '0')

        self.add_text_element(form_document, voices_element, "count", "count", str(len(self.xml_form_dict['voices'])))
        self.add_text_element(form_document, voices_element, "item_version", "item_version", "0")

        write_version = True
        for v in self.xml_form_dict['voices']:
            self.add_voice_element(form_document, voices_element, v, write_version)
            write_version = False

        xml_form_element.appendChild(voices_element)

        form_document.writexml(open(afilename, 'w'), indent="", addindent="\t", newl='\n', encoding="UTF-8", standalone=True)
        form_document.unlink()


    def add_voice_element(self, doc, parent, voice_dict, write_version):
        item_element = doc.createElement("item")
        if (write_version):
            item_element.setAttribute("class_id", '6')
            item_element.setAttribute("tracking_level", '0')
            item_element.setAttribute("version", '0')

        self.add_text_element(doc, item_element, "low_pitch", "low_pitch_", str(voice_dict['low_pitch']))
        self.add_text_element(doc, item_element, "high_pitch", "high_pitch_", str(voice_dict['high_pitch']))
        self.add_text_element(doc, item_element, "channel", "channel_", str(int(voice_dict['channel'])))
        self.add_text_element(doc, item_element, "walking", "walking_", str(int(voice_dict['walking'])))
        self.add_text_element(doc, item_element, "program", "program_", str(voice_dict['program']))
        self.add_text_element(doc, item_element, "pan", "pan_", str(int(voice_dict['pan'])))
        follower_element = doc.createElement("follower_")
        if (write_version):
            follower_element.setAttribute("class_id", '7')
            follower_element.setAttribute("tracking_level", '0')
            follower_element.setAttribute("version", '0')

        self.add_text_element(doc, follower_element, "follow", "follow_", str(int(voice_dict['follower']['follow'])))
        self.add_text_element(doc, follower_element, "leader", "leader_", str(int(voice_dict['follower']['leader'])))
        self.add_text_element(doc, follower_element, "interval_type", "interval_type_", str(int(voice_dict['follower']['interval_type'])))
        self.add_text_element(doc, follower_element, "interval", "interval_", str(int(voice_dict['follower']['interval'])))
        item_element.appendChild(follower_element)
        parent.appendChild(item_element)

    def add_form_element(self, doc, parent, form_name, add_attributes = False):
        form_element = doc.createElement(form_name + "_")
        if (add_attributes == True):
            form_element.setAttribute("class_id", '3')
            form_element.setAttribute("tracking_level", '0')
            form_element.setAttribute("version", '0')

        mean_sine_element = doc.createElement("mean_sine_")
        if (add_attributes == True):
            mean_sine_element.setAttribute("class_id", '4')
            mean_sine_element.setAttribute("tracking_level", '0')
            mean_sine_element.setAttribute("version", '0')
        self.add_text_element(doc, mean_sine_element , "period", "period_", str(self.xml_form_dict[form_name]['mean']['period']))
        self.add_text_element(doc, mean_sine_element , "phase", "phase_", str(self.xml_form_dict[form_name]['mean']['phase']))
        self.add_text_element(doc, mean_sine_element , "amplitude", "amplitude_", str(self.xml_form_dict[form_name]['mean']['amplitude']))
        self.add_text_element(doc, mean_sine_element , "offset", "offset_", str(self.xml_form_dict[form_name]['mean']['offset']))
        form_element.appendChild(mean_sine_element)
        range_sine_element = doc.createElement("range_sine_")
        self.add_text_element(doc, range_sine_element , "period", "period_", str(self.xml_form_dict[form_name]['range']['period']))
        self.add_text_element(doc, range_sine_element , "phase", "phase_", str(self.xml_form_dict[form_name]['range']['phase']))
        self.add_text_element(doc, range_sine_element , "amplitude", "amplitude_", str(self.xml_form_dict[form_name]['range']['amplitude']))
        self.add_text_element(doc, range_sine_element , "offset", "offset_", str(self.xml_form_dict[form_name]['range']['offset']))
        form_element.appendChild(range_sine_element)
        parent.appendChild(form_element)

    def add_sine_element(self, doc, parent, sine_name):
        sine_element = doc.createElement(sine_name + "_")
        self.add_text_element(doc, sine_element , "period", "period_", str(self.xml_form_dict[sine_name]['period']))
        self.add_text_element(doc, sine_element , "phase", "phase_", str(self.xml_form_dict[sine_name]['phase']))
        self.add_text_element(doc, sine_element , "amplitude", "amplitude_", str(self.xml_form_dict[sine_name]['amplitude']))
        self.add_text_element(doc, sine_element , "offset", "offset_", str(self.xml_form_dict[sine_name]['offset']))
        parent.appendChild(sine_element)

    def add_text_element(self, doc, parent, name, xmltag = None, value = None):
        if (xmltag == None):
            xmltag = name + "_"
        if (value == None):
            value = str(self.xml_form_dict[name])

        any_element = doc.createElement(xmltag)
        parent.appendChild(any_element)
        any_text_node = doc.createTextNode(value)
        any_element.appendChild(any_text_node)
        parent.appendChild(any_element)

    def redraw_callback(self):
        self.xml_form_dict['name'] = self.all_forms_window.xml_form['name']
        self.xml_form_dict['len'] = self.all_forms_window.xml_form['len']
        self.xml_form_dict['min_note_len'] = self.all_forms_window.xml_form['min_note_len']
        self.xml_form_dict['max_note_len'] = self.all_forms_window.xml_form['max_note_len']
        self.xml_form_dict['pulse'] = self.all_forms_window.xml_form['pulse']
        self.xml_form_dict['melody_probabilities']['down'] = self.all_forms_window.xml_form['melody_probabilities']['down']
        self.xml_form_dict['melody_probabilities']['same'] = self.all_forms_window.xml_form['melody_probabilities']['same']
        self.xml_form_dict['melody_probabilities']['up'] = self.all_forms_window.xml_form['melody_probabilities']['up']

        self.xml_form_dict['pitch_form']['mean']['period'] = self.all_forms_window.xml_form['pitch_form']['mean']['period']
        self.xml_form_dict['pitch_form']['mean']['phase'] = self.all_forms_window.xml_form['pitch_form']['mean']['phase']
        self.xml_form_dict['pitch_form']['mean']['amplitude'] = self.all_forms_window.xml_form['pitch_form']['mean']['amplitude']
        self.xml_form_dict['pitch_form']['mean']['offset'] = self.all_forms_window.xml_form['pitch_form']['mean']['offset']

        self.xml_form_dict['rhythm_form']['mean']['period'] = self.all_forms_window.xml_form['rhythm_form']['mean']['period']
        self.xml_form_dict['rhythm_form']['mean']['phase'] = self.all_forms_window.xml_form['rhythm_form']['mean']['phase']
        self.xml_form_dict['rhythm_form']['mean']['amplitude'] = self.all_forms_window.xml_form['rhythm_form']['mean']['amplitude']
        self.xml_form_dict['rhythm_form']['mean']['offset'] = self.all_forms_window.xml_form['rhythm_form']['mean']['offset']

        self.xml_form_dict['dynamic_form']['mean']['period'] = self.all_forms_window.xml_form['dynamic_form']['mean']['period']
        self.xml_form_dict['dynamic_form']['mean']['phase'] = self.all_forms_window.xml_form['dynamic_form']['mean']['phase']
        self.xml_form_dict['dynamic_form']['mean']['amplitude'] = self.all_forms_window.xml_form['dynamic_form']['mean']['amplitude']
        self.xml_form_dict['dynamic_form']['mean']['offset'] = self.all_forms_window.xml_form['dynamic_form']['mean']['offset']

        self.xml_form_dict['texture_form']['period'] = self.all_forms_window.xml_form['texture_form']['period']
        self.xml_form_dict['texture_form']['phase'] = self.all_forms_window.xml_form['texture_form']['phase']
        self.xml_form_dict['texture_form']['amplitude'] = self.all_forms_window.xml_form['texture_form']['amplitude']
        self.xml_form_dict['texture_form']['offset'] = self.all_forms_window.xml_form['texture_form']['offset']

        self.xml_form_dict['pitch_form']['range']['period'] = self.all_forms_window.xml_form['pitch_form']['range']['period']
        self.xml_form_dict['pitch_form']['range']['phase'] = self.all_forms_window.xml_form['pitch_form']['range']['phase']
        self.xml_form_dict['pitch_form']['range']['amplitude'] = self.all_forms_window.xml_form['pitch_form']['range']['amplitude']
        self.xml_form_dict['pitch_form']['range']['offset'] = self.all_forms_window.xml_form['pitch_form']['range']['offset']

        self.xml_form_dict['rhythm_form']['range']['period'] = self.all_forms_window.xml_form['rhythm_form']['range']['period']
        self.xml_form_dict['rhythm_form']['range']['phase'] = self.all_forms_window.xml_form['rhythm_form']['range']['phase']
        self.xml_form_dict['rhythm_form']['range']['amplitude'] = self.all_forms_window.xml_form['rhythm_form']['range']['amplitude']
        self.xml_form_dict['rhythm_form']['range']['offset'] = self.all_forms_window.xml_form['rhythm_form']['range']['offset']

        self.xml_form_dict['dynamic_form']['range']['period'] = self.all_forms_window.xml_form['dynamic_form']['range']['period']
        self.xml_form_dict['dynamic_form']['range']['phase'] = self.all_forms_window.xml_form['dynamic_form']['range']['phase']
        self.xml_form_dict['dynamic_form']['range']['amplitude'] = self.all_forms_window.xml_form['dynamic_form']['range']['amplitude']
        self.xml_form_dict['dynamic_form']['range']['offset'] = self.all_forms_window.xml_form['dynamic_form']['range']['offset']

        self.draw_form()

    def about_callback(self):
        about_top = tkinter.Tk()
        about_window = tkinter.Text(about_top)
        about_window.grid(sticky="we", row=0, column=0)
        about_top.title('About')
        about_window.insert('1.0', "textmidiformedit musical form editor by Thomas E. Janzen\n2021\nUse with textmidicgm, part of TextMIDITools\nat github.com/tomejanzen/TextMIDITools")
        about_window['state'] = "disabled"
        about_window.focus()

def textmidiformedit():
    xmlform_window = XmlForm()
    xmlform_window.frame.grid(sticky="we", row=0, column=0)
    xmlform_window.frame.rowconfigure(index=0, weight=1)
    xmlform_window.frame.columnconfigure(index=0, weight=1)

    xmlform_window.mainloop()

if __name__ == '__main__':
    textmidiformedit()




#!/usr/bin/env python3
# TextMIDITools Version 1.0.18
# textmidiform.py 1.0
# Copyright © 2022 Thomas E. Janzen
# License GPLv3+: GNU GPL version 3 or later <https://gnu.org/licenses/gpl.html>
# This is free software: you are free to change and redistribute it.
# There is NO WARRANTY, to the extent permitted by law.
#
import tkinter
import tkinter.ttk
from tkinter import *
from tkinter.ttk import *
import re

class Sine(tkinter.Frame):
    pi = 3.1415926535
    def __init__(self, parent, sine_title, period_callback, phase_callback, xml_sine):
        # assignment of class data must be first
        self.the_row = 0
        self.period_callback = period_callback
        self.phase_callback = phase_callback
        super().__init__(parent)
        self.sine_title = sine_title
        self.xml_sine = xml_sine
        self.grid(sticky=NSEW, padx=100)
        self.create_widgets()

    def create_widgets(self):

        self.curve_label = tkinter.ttk.Label(self, text=self.sine_title)
        self.curve_label.grid(row=1, column=0)
        self.rowconfigure(index=self.the_row, weight=1)
        self.period_label = tkinter.ttk.Label(self, text="Period")
        self.period_label.grid(row=self.the_row, column=1, sticky=NSEW)
        # TCL book p 509
        # %d type 1 insert, 0 del, -1 other
        # %i index of string to change or -1
        # %P value of entry if edit is allowed
        # %s val of entry prior to editing
        # %S the text string being inserted or deleted
        # %v type of validation currently set
        # %V type of validation that triggered us (key, focusin, focusout, forced)
        # %W tk name of widget
        validate_command = (self.register(self.validate_length),
            '%d', '%i', '%P', '%s', '%S', '%v', '%V', '%W')
        self.period_entry = tkinter.ttk.Entry(self, validatecommand=validate_command, validate="focusout")
        self.period_entry.delete(0, 1024)
        self.period_entry.insert(0, self.xml_sine['period'])
        self.period_entry.update()
        self.period_entry.grid(row=self.the_row, column=2, sticky=NSEW)
        self.period_entry.bind('<Key-Return>', self.period_callback)
        self.period_entry.bind('<FocusOut>', self.period_callback)
        self.the_row = self.the_row + 1

        self.rowconfigure(index=self.the_row, weight=1)
        self.phase_scroll_label = tkinter.ttk.Label(self, text="-pi..................0..................pi")
        self.phase_scroll_label.rowconfigure(index=self.the_row, weight=1)
        self.phase_scroll_label.grid(row=self.the_row, column=2, sticky=NSEW)
        self.the_row = self.the_row + 1

        self.rowconfigure(index=self.the_row, weight=1)
        self.phase_label = tkinter.ttk.Label(self, text="Phase")
        self.phase_label.rowconfigure(index=self.the_row, weight=1)
        self.phase_label.grid(row=self.the_row, column=1, sticky=NSEW)
        self.phase_scrollbar = tkinter.ttk.Scrollbar(self, command=self.phase_callback)
        self.phase_scrollbar["orient"] = "horizontal"
        self.phase_scrollbar.grid(row=self.the_row, column=2, sticky=NSEW)
        self.phase_scrollbar.rowconfigure(index=self.the_row, weight=1)
        phase = self.xml_sine['phase'] / (2.0 * self.pi) + 0.5
        self.phase_scrollbar.set(phase, phase)
        self.the_row = self.the_row + 1

    def install_xml_sine(self, xml_sine):
        self.xml_sine = xml_sine
        self.period_entry.delete(0, 1024)
        self.period_entry.insert(0, self.xml_sine['period'])
        self.period_entry.update()
        phase = self.xml_sine['phase'] / (2.0 * self.pi) + 0.5
        self.phase_scrollbar.set(phase, phase)
        self.phase_scrollbar.update()
        self.update()
    def validate_length(self, d, i, P, s, S, v, V, W):
        # 1-insert, 0=del, -1 if forced
        pat = re.compile('\d+([.]\d*)?')
        ma = pat.match(str(P))
        ret = False
        if (ma):
            ret = True
        else:
            ret = False
        return ret




#!/usr/bin/env python3
# TextMIDITools Version 1.0.18
# textmidiform.py 1.0
# Copyright © 2022 Thomas E. Janzen
# License GPLv3+: GNU GPL version 3 or later <https://gnu.org/licenses/gpl.html>
# This is free software: you are free to change and redistribute it.
# There is NO WARRANTY, to the extent permitted by law.
#
import math
import tkinter
import tkinter.ttk
from tkinter import *
from tkinter import ttk
from tkinter.ttk import *
import tkinter, tkinter.constants, tkinter.filedialog
import re

from FormFrame import FormFrame
from Sine import Sine
from KeyboardWindow import KeyboardWindow

class ScaleFrame(tkinter.Frame):
    scale_names = []
    scale_steps_dict = {'Unnamed'    : [0, 2, 4, 5, 7, 9, 11],
                        'Diatonic'   : [0, 2, 4, 5, 7, 9, 11],
                        'Minor'      : [0, 2, 3, 5, 7, 8, 11],
                        'WholeTone'  : [0, 2, 4, 6, 8, 10],
                        'Diminished' : [0, 2, 3, 5, 6, 8, 9, 11],
                        'Chromatic'  : [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11],
                        'Pentatonic' : [0, 2, 5, 7, 9],
                        'Tritone' : [0, 2, 4, 5, 6, 8, 10, 11]}
    full_midi_scale = ["C-1", "C#-1", "D-1", "D#-1", "E-1", "F-1", "F#-1", "G-1", "G#-1", "A-1", "A#-1", "B-1",
                       "C0", "C#0", "D0", "D#0", "E0", "F0", "F#0", "G0", "G#0", "A0", "A#0", "B0",
                       "C1", "C#1", "D1", "D#1", "E1", "F1", "F#1", "G1", "G#1", "A1", "A#1", "B1",
                       "C2", "C#2", "D2", "D#2", "E2", "F2", "F#2", "G2", "G#2", "A2", "A#2", "B2",
                       "C3", "C#3", "D3", "D#3", "E3", "F3", "F#3", "G3", "G#3", "A3", "A#3", "B3",
                       "C4", "C#4", "D4", "D#4", "E4", "F4", "F#4", "G4", "G#4", "A4", "A#4", "B4",
                       "C5", "C#5", "D5", "D#5", "E5", "F5", "F#5", "G5", "G#5", "A5", "A#5", "B5",
                       "C6", "C#6", "D6", "D#6", "E6", "F6", "F#6", "G6", "G#6", "A6", "A#6", "B6",
                       "C7", "C#7", "D7", "D#7", "E7", "F7", "F#7", "G7", "G#7", "A7", "A#7", "B7",
                       "C8", "C#8", "D8", "D#8", "E8", "F8", "F#8", "G8", "G#8", "A8", "A#8", "B8",
                       "C9", "C#9", "D9", "D#9", "E9", "F9", "F#9", "G9"]
    notenames = {'C' : 0, 'D' : 2, 'E' : 4, 'F' : 5, 'G' : 7, 'A' : 9, 'B' : 11}
    accidentals = {'b' : -1, '#' : 1, 'bb' : -2, 'x' : 2}
    half_steps_per_octave = 12
    midi_keys_qty = 128
    midi_octaves_qty = 11

    def __init__(self, topwin):
        super().__init__(topwin)
        self.grid(row=0, padx=10, pady=10)
        self.scale_names = list(self.scale_steps_dict)
        self.scales_dict = {}
        self.scale_untransposed = []
        for sn in self.scale_names:
            self.scales_dict[sn] = []
            for octave in range(0, self.midi_octaves_qty):
                for step in self.scale_steps_dict[sn]:
                    key = octave * self.half_steps_per_octave + step
                    if (key < self.midi_keys_qty):
                        self.scales_dict[sn].append(key)
        self.scale_transposed = []
        for k in self.scales_dict['Chromatic'] :
            self.scale_untransposed.append(k)
            self.scale_transposed.append(k)
        self.transpose_num = 0
        self.create_widgets()

    def set_keyboard(self):
        midi_key_selects = []
        for s in range(0, len(self.full_midi_scale)):
            midi_key_selects.append(False)
        for k in self.scale_transposed:
            midi_key_selects[k] = True
        self.keyboard_window.keyboard_frame.set_selects(midi_key_selects)

    def set_keyboard_from_xml(self, xml_form):
        scale_ints = []
        for key_string in xml_form['scale']:
            key_string[:1].upper()
            notename = key_string[0]
            if (len(key_string) == 2): # C0
                keynum = ScaleFrame.notenames[notename] + ScaleFrame.half_steps_per_octave * (int(key_string[1]) + 1)
            if (len(key_string) == 3): # C-1 Cb1 C#1 Cx2
                if (key_string[1:3] == '-1'): # bottom octave
                    keynum = ScaleFrame.notenames[notename]
                else: # C#1
                    keynum = ScaleFrame.notenames[notename] + ScaleFrame.accidentals[key_string[1]] + ScaleFrame.half_steps_per_octave * (int(key_string[2]) + 1)
            if (len(key_string) == 4): # Db-1 Cbb1 C#-1 Ebb4
                if (key_string[2:4] == '-1'):
                    keynum = ScaleFrame.notenames[notename] + ScaleFrame.accidentals[key_string[1]]
                else:
                    if (key_string[1:3] == 'bb'):
                        keynum = ScaleFrame.notenames[notename] + ScaleFrame.accidentals[key_string[1:3]] + ScaleFrame.half_steps_per_octave * (int(key_string[3]) + 1)
            if (len(key_string) == 5): # Dbb-1 ;there is no Cbb-1
                if ((key_string[3:5] == '-1') and (key_string[1:3] == 'bb')):
                    keynum = ScaleFrame.notenames[notename] + ScaleFrame.accidentals[key_string[1:3]]
            scale_ints.append(keynum)

        midi_key_selects = []

        for s in range(0, len(ScaleFrame.full_midi_scale)):
            midi_key_selects.append(False)
        for si in scale_ints:
            midi_key_selects[si] = True
        self.keyboard_window.keyboard_frame.set_selects(midi_key_selects)
        self.scale_untransposed = []
        for k in range(0, len(midi_key_selects)):
            if (midi_key_selects[k]):
                self.scale_untransposed.append(k)
        self.transpose_scale()

    def create_widgets(self):
        the_row = 0

        self.keyboard_window = KeyboardWindow()
        self.keyboard_window.geometry('1500x162+50-50')
        self.set_keyboard()

        self.rowconfigure(index=the_row, weight=1)
        self.columnconfigure(index=0, weight=1)
        self.columnconfigure(index=1, weight=1)
        self.scale_name_label = tkinter.ttk.Label(self, text="Scale")
        self.scale_name_label.grid(row=the_row, column=0, sticky=NSEW)
        self.scale_name_spinbox = tkinter.ttk.Spinbox(self, wrap=True)
        self.scale_name_spinbox["values"] = self.scale_names
        self.scale_name_spinbox.grid(row=the_row, column=1, sticky=NSEW)
        self.scale_name_spinbox.bind('<ButtonRelease-1>', self.scale_name_callback)
        self.scale_name_spinbox.set("Unnamed")
        self.scale_name_spinbox['state'] = 'readonly'
        the_row = the_row + 1

        self.rowconfigure(index=the_row, weight=1)
        self.transpose_label = tkinter.ttk.Label(self, text="Transpose")
        self.transpose_label.grid(row=the_row, column=0, sticky=NSEW)
        self.transpose_spinbox = tkinter.ttk.Spinbox(self, wrap=True)
        self.transpose_spinbox["increment"] = 1
        self.transpose_spinbox["from"]      = -5
        self.transpose_spinbox["to"]        = 6
        self.transpose_spinbox.grid(row=the_row, column=1, sticky=NSEW)
        self.transpose_spinbox.bind('<ButtonRelease-1>', self.transpose_callback)
        self.transpose_spinbox.set(0)
        self.transpose_spinbox['state'] = 'readonly'
        self.from_keyboard_button = tkinter.ttk.Button(self, text='From Keyboard',
            command=self.from_keyboard_callback)
        self.from_keyboard_button.grid(row=the_row, column=2, stick=NSEW)
        the_row = the_row + 1

    def from_keyboard_callback(self):
        key_ints = self.keyboard_window.keyboard_frame.get_key_ints()
        self.scale_untransposed = []
        for key in key_ints:
            self.scale_untransposed.append(key)
        self.transpose_scale()
        self.scale_name_spinbox['state'] = '!readonly'
        self.scale_name_spinbox.set(self.scale_names[0])
        self.scale_name_spinbox['state'] = 'readonly'

        self.transpose_spinbox['state'] = '!readonly'
        self.transpose_spinbox.set(0)
        self.transpose_spinbox['state'] = 'readonly'

    def transpose_callback(self, event):
        self.transpose_num = int(self.transpose_spinbox.get())
        self.transpose_scale()
        self.set_keyboard()

    def scale_name_callback(self, event):
        self.scale_name_spinbox.update()
        scale_name = self.scale_name_spinbox.get()
        self.scale_untransposed = []
        for keynum in self.scales_dict[scale_name]:
            self.scale_untransposed.append(keynum)
        self.transpose_scale()
        self.set_keyboard()

    def transpose_scale(self):
        self.scale_transposed = []
        for keynum in self.scale_untransposed:
            key_index = keynum + self.transpose_num
            self.scale_transposed.append(key_index)

        for k in range(0, len(self.scale_transposed)):
            if (self.scale_transposed[k] < 0):
                self.scale_transposed[k] = self.scale_transposed[k] + 128

        for k in range(0, len(self.scale_transposed)):
            if (self.scale_transposed[k]  >= self.midi_keys_qty):
                self.scale_transposed[k]  = self.scale_transposed[k]  - 128

        low = 0
        for k in range(1, len(self.scale_transposed)):
            if (self.scale_transposed[k - 1] > self.scale_transposed[k]):
                low = k
        right_partition = self.scale_transposed[0:low]
        left_partition  = self.scale_transposed[low:]
        left_partition.extend(right_partition)
        self.scale_transposed = []
        for key in left_partition:
            self.scale_transposed.append(key)

    def get_string_scale(self):
        scale_strings = []
        for keynum in self.scale_transposed:
            scale_strings.append(self.full_midi_scale[keynum])
        return scale_strings

class AllFormsWindow(tkinter.Toplevel):
    twopi = 2.0 * math.pi
    def __init__(self, xml_form = None):
        super().__init__()
        self.frame = ttk.Frame(self, padding="1 1 1 1")
        self.xml_form = xml_form
        self.create_widgets()
        self.frame.grid(sticky="nw", row=0, column=0)
        self.frame.rowconfigure(index=0, weight=1)
        self.frame.columnconfigure(index=0, weight=1)
        self.title('Musical Form')
        self.geometry('400x850+20+50')

    def create_widgets(self):
        the_row = 0

        self.frame.rowconfigure(index=the_row, weight=1)
        self.name_label = tkinter.ttk.Label(self.frame, text="Name")
        self.name_label.grid(row=the_row, column=0, sticky=NSEW)
        self.name_entry = tkinter.ttk.Entry(self.frame)
        self.name_entry.insert(0, self.xml_form['name'])
        self.name_entry.grid(row=the_row, column=1, sticky=NSEW)
        self.name_entry.bind('<Key-Return>', self.name_callback)
        self.name_entry.bind('<FocusOut>', self.name_callback)
        the_row = the_row + 1

        self.frame.rowconfigure(index=the_row, weight=1)
        self.frame.columnconfigure(index=0, weight=1)
        self.frame.columnconfigure(index=1, weight=1)
        self.len_label = tkinter.ttk.Label(self.frame, text="Len")
        self.len_label.grid(row=the_row, column=0, sticky=NSEW)
        self.len_entry = tkinter.ttk.Entry(self.frame)
        self.len_entry.insert(0, self.xml_form['len'])
        self.len_entry.grid(row=the_row, column=1, sticky=NSEW)
        self.len_entry.bind('<Key-Return>', self.len_callback)
        self.len_entry.bind('<FocusOut>', self.len_callback)
        the_row = the_row + 1

        self.frame.rowconfigure(index=the_row, weight=1)
        self.min_note_len_label = tkinter.ttk.Label(self.frame, text="Min Note Len")
        self.min_note_len_label.grid(row=the_row, column=0, sticky=NSEW)
        self.min_note_len_entry = tkinter.ttk.Entry(self.frame)
        self.min_note_len_entry.insert(0, self.xml_form['min_note_len'])
        self.min_note_len_entry.grid(row=the_row, column=1, sticky=NSEW)
        self.min_note_len_entry.bind('<Key-Return>', self.min_note_len_callback)
        self.min_note_len_entry.bind('<FocusOut>', self.min_note_len_callback)
        the_row = the_row + 1

        self.frame.rowconfigure(index=the_row, weight=1)
        self.max_note_len_label = tkinter.ttk.Label(self.frame, text="Max Note Len")
        self.max_note_len_label.grid(row=the_row, column=0, sticky=NSEW)
        self.max_note_len_entry = tkinter.ttk.Entry(self.frame)
        self.max_note_len_entry.insert(0, self.xml_form['max_note_len'])
        self.max_note_len_entry.grid(row=the_row, column=1, sticky=NSEW)
        self.max_note_len_entry.bind('<Key-Return>', self.max_note_len_callback)
        self.max_note_len_entry.bind('<FocusOut>', self.max_note_len_callback)
        the_row = the_row + 1

        self.frame.rowconfigure(index=the_row, weight=1)
        self.pulse_label = tkinter.ttk.Label(self.frame, text="Pulse/Sec")
        self.pulse_label.grid(row=the_row, column=0, sticky=NSEW)
        self.pulse_entry = tkinter.ttk.Entry(self.frame)
        self.pulse_entry.insert(0, self.xml_form['pulse'])
        self.pulse_entry.grid(row=the_row, column=1, sticky=NSEW)
        self.pulse_entry.bind('<Key-Return>', self.pulse_callback)
        self.pulse_entry.bind('<FocusOut>', self.pulse_callback)
        the_row = the_row + 1

        self.frame.rowconfigure(index=the_row, weight=1)
        self.melody_probabilities_label = tkinter.ttk.Label(self.frame, text="Melody Probabilities")
        self.melody_probabilities_label.grid(row=the_row, column=0, sticky=NSEW)
        self.melody_probabilities_entry = tkinter.ttk.Entry(self.frame)

        self.frame.rowconfigure(index=the_row, weight=1)
        self.down_label = tkinter.ttk.Label(self.frame, text="Down")
        self.down_label.grid(row=the_row, column=0, sticky=NSEW)
        self.down_entry = tkinter.ttk.Entry(self.frame)
        self.down_entry.insert(0, self.xml_form['melody_probabilities']['down'])
        self.down_entry.grid(row=the_row, column=1, sticky=NSEW)
        self.down_entry.bind('<Key-Return>', self.melody_down_callback)
        self.down_entry.bind('<FocusOut>', self.melody_down_callback)
        the_row = the_row + 1

        self.frame.rowconfigure(index=the_row, weight=1)
        self.same_label = tkinter.ttk.Label(self.frame, text="Same")
        self.same_label.grid(row=the_row, column=0, sticky=NSEW)
        self.same_entry = tkinter.ttk.Entry(self.frame)
        self.same_entry.insert(0, self.xml_form['melody_probabilities']['same'])
        self.same_entry.grid(row=the_row, column=1, sticky=NSEW)
        self.same_entry.bind('<Key-Return>', self.melody_same_callback)
        self.same_entry.bind('<FocusOut>', self.melody_same_callback)
        the_row = the_row + 1

        self.frame.rowconfigure(index=the_row, weight=1)
        self.up_label = tkinter.ttk.Label(self.frame, text="Up")
        self.up_label.grid(row=the_row, column=0, sticky=NSEW)
        self.up_entry = tkinter.ttk.Entry(self.frame)
        self.up_entry.insert(0, self.xml_form['melody_probabilities']['up'])
        self.up_entry.grid(row=the_row, column=1, sticky=NSEW)
        self.up_entry.bind('<Key-Return>', self.melody_up_callback)
        self.up_entry.bind('<FocusOut>', self.melody_up_callback)
        the_row = the_row + 1

        self.scale_frame = ScaleFrame(self)
        self.scale_frame.grid(sticky="w", row=the_row, column=0)
        self.scale_frame.rowconfigure(index=0, weight=1)
        self.scale_frame.columnconfigure(index=0, weight=1)

        the_row = the_row + 2

        the_row = the_row + 1
        self.pitch_form = FormFrame(parent=self, form_title="PITCH", mean_period_callback=self.pitch_mean_period_callback, mean_phase_callback=self.pitch_mean_phase_callback, range_period_callback=self.pitch_range_period_callback, range_phase_callback=self.pitch_range_phase_callback, xml_subform=self.xml_form['pitch_form'])
        self.pitch_form.grid(row=the_row, column=0, sticky=NSEW)
        the_row = the_row + 1

        the_row = the_row + 1
        self.rhythm_form = FormFrame(parent=self, form_title="RHYTHM", mean_period_callback=self.rhythm_mean_period_callback, mean_phase_callback=self.rhythm_mean_phase_callback, range_period_callback=self.rhythm_range_period_callback, range_phase_callback=self.rhythm_range_phase_callback, xml_subform=self.xml_form['rhythm_form'])
        self.rhythm_form.grid(row=the_row, column=0, sticky=NSEW)
        the_row = the_row + 1

        the_row = the_row + 1
        self.dynamic_form = FormFrame(parent=self, form_title="DYNAMIC", mean_period_callback=self.dynamic_mean_period_callback, mean_phase_callback=self.dynamic_mean_phase_callback, range_period_callback=self.dynamic_range_period_callback, range_phase_callback=self.dynamic_range_phase_callback, xml_subform=self.xml_form['dynamic_form'])
        self.dynamic_form.grid(row=the_row, column=0, sticky=NSEW)
        the_row = the_row + 1

        the_row = the_row + 1
        self.texture_form = Sine(parent=self, sine_title="TEXTURE", period_callback=self.texture_range_period_callback, phase_callback=self.texture_range_phase_callback, xml_sine=self.xml_form['texture_form'])
        self.texture_form.grid(row=the_row, column=0, sticky=NSEW)

    def name_callback(self, event):
        self.xml_form['name'] = self.name_entry.get()

    def len_callback(self, event):
        self.xml_form['len'] = self.len_entry.get()

    def min_note_len_callback(self, event):
        self.xml_form['min_note_len'] = self.min_note_len_entry.get()

    def max_note_len_callback(self, event):
        self.xml_form['max_note_len'] = self.max_note_len_entry.get()

    def pulse_callback(self, event):
        self.xml_form['pulse'] = self.pulse_entry.get()

    def melody_down_callback(self, event):
        self.xml_form['melody_probabilities']['down'] = self.down_entry.get()

    def melody_same_callback(self, event):
        self.xml_form['melody_probabilities']['same'] = self.same_entry.get()

    def melody_up_callback(self, event):
        self.xml_form['melody_probabilities']['up'] = self.up_entry.get()

    def pitch_mean_period_callback(self, event):
        self.xml_form['pitch_form']['mean']['period'] = self.pitch_form.mean.period_entry.get()

    def pitch_mean_phase_callback(self, action, sign, unitspage=None):
        self.pitch_form.mean.phase_scrollbar.update()
        firstlastlist = self.pitch_form.mean.phase_scrollbar.get()
        if (action == "scroll"):
            if (unitspage == "pages"):
                delta = float(sign) * 1.0 / 4
                setting = firstlastlist[0]
                if ((setting <= 0.0) and (delta < 0.0)):
                    delta = 0.0
                if ((setting >= 1.0) and (delta > 0.0)):
                    delta = 0.0
                setting = setting + delta
                self.pitch_form.mean.phase_scrollbar.set(setting, setting)
                value = (setting - 0.5) * self.twopi
                self.xml_form['pitch_form']['mean']['phase'] = value
            else:
                if (unitspage == "units"):
                    delta = float(sign) * 1.0 / 32
                    setting = firstlastlist[0]
                    if ((setting <= 0.0) and (delta < 0.0)):
                        delta = 0.0
                    if ((setting >= 1.0) and (delta > 0.0)):
                        delta = 0.0
                    setting = setting + delta
                    self.pitch_form.mean.phase_scrollbar.set(setting, setting)
                    value = (setting - 0.5) * self.twopi
                    self.xml_form['pitch_form']['mean']['phase'] = value
        else:
            if (action == "goto"):
                setting = firstlastlist[0]
                self.pitch_form.mean.phase_scrollbar.set(setting)
                value = (setting - 0.5) * self.twopi
                self.xml_form['pitch_form']['mean']['phase'] = value

    def pitch_range_period_callback(self, event):
        self.xml_form['pitch_form']['range']['period'] = self.pitch_form.range.period_entry.get()

    def pitch_range_phase_callback(self, action, sign, unitspage=None):
        firstlastlist = self.pitch_form.range.phase_scrollbar.get()
        if (action == "scroll"):
            if (unitspage == "pages"):
                delta = float(sign) * 1.0 / 4
                setting = firstlastlist[0]
                if ((setting <= 0.0) and (delta < 0.0)):
                    delta = 0.0
                if ((setting >= 1.0) and (delta > 0.0)):
                    delta = 0.0
                setting = setting + delta
                self.pitch_form.range.phase_scrollbar.set(setting, setting)
                value = (setting - 0.5) * self.twopi
                self.xml_form['pitch_form']['range']['phase'] = value
            else:
                if (unitspage == "units"):
                    delta = float(sign) * 1.0 / 32
                    setting = firstlastlist[0]
                    if ((setting <= 0.0) and (delta < 0.0)):
                        delta = 0.0
                    if ((setting >= 1.0) and (delta > 0.0)):
                        delta = 0.0
                    setting = setting + delta
                    self.pitch_form.range.phase_scrollbar.set(setting, setting)
                    value = (setting - 0.5) * self.twopi
                    self.xml_form['pitch_form']['range']['phase'] = value
        else:
            if (action == "goto"):
                setting = firstlastlist[0]
                self.pitch_form.range.phase_scrollbar.set(setting, setting)
                value = (setting - 0.5) * self.twopi
                self.xml_form['pitch_form']['range']['phase'] = value

    def rhythm_mean_period_callback(self, event):
        self.xml_form['rhythm_form']['mean']['period'] = self.rhythm_form.mean.period_entry.get()

    def rhythm_mean_phase_callback(self, action, sign, unitspage):
        firstlastlist = self.rhythm_form.mean.phase_scrollbar.get()
        if (action == "scroll"):
            if (unitspage == "pages"):
                delta = float(sign) * 1.0 / 4
                setting = firstlastlist[0]
                if ((setting <= 0.0) and (delta < 0.0)):
                    delta = 0.0
                if ((setting >= 1.0) and (delta > 0.0)):
                    delta = 0.0
                setting = setting + delta
                self.rhythm_form.mean.phase_scrollbar.set(setting, setting)
                value = (setting - 0.5) * self.twopi
                self.xml_form['rhythm_form']['mean']['phase'] = value
            else:
                if (unitspage == "units"):
                    delta = float(sign) * 1.0 / 32
                    setting = firstlastlist[0]
                    if ((setting <= 0.0) and (delta < 0.0)):
                        delta = 0.0
                    if ((setting >= 1.0) and (delta > 0.0)):
                        delta = 0.0
                    setting = setting + delta
                    self.rhythm_form.mean.phase_scrollbar.set(setting, setting)
                    value = (setting - 0.5) * self.twopi
                    self.xml_form['rhythm_form']['mean']['phase'] = value
        else:
            if (action == "goto"):
                setting = firstlastlist[0]
                value = (setting - 0.5) * self.twopi
                self.pitch_form.range.phase_scrollbar.set(setting, setting)
                self.xml_form['rhythm_form']['mean']['phase'] = value

    def rhythm_range_period_callback(self, event):
        self.xml_form['rhythm_form']['range']['period'] = self.rhythm_form.range.period_entry.get()

    def rhythm_range_phase_callback(self, action, sign, unitspage=None):
        firstlastlist = self.rhythm_form.range.phase_scrollbar.get()
        if (action == "scroll"):
            if (unitspage == "pages"):
                delta = float(sign) * 1.0 / 4
                setting = firstlastlist[0]
                if ((setting <= 0.0) and (delta < 0.0)):
                    delta = 0.0
                if ((setting >= 1.0) and (delta > 0.0)):
                    delta = 0.0
                setting = setting + delta
                self.rhythm_form.range.phase_scrollbar.set(setting, setting)
                value = (setting - 0.5) * self.twopi
                self.xml_form['rhythm_form']['range']['phase'] = value
            else:
                if (unitspage == "units"):
                    delta = float(sign) * 1.0 / 32
                    setting = firstlastlist[0]
                    if ((setting <= 0.0) and (delta < 0.0)):
                        delta = 0.0
                    if ((setting >= 1.0) and (delta > 0.0)):
                        delta = 0.0
                    setting = setting + delta
                    self.rhythm_form.range.phase_scrollbar.set(setting, setting)
                    value = (setting - 0.5) * self.twopi
                    self.xml_form['rhythm_form']['range']['phase'] = value
        else:
            if (action == "goto"):
                setting = firstlastlist[0]
                self.rhythms_form.range.phase_scrollbar.set(setting, setting)
                value = (setting - 0.5) * self.twopi
                self.xml_form['rhythm_form']['range']['phase'] = value

    def dynamic_mean_period_callback(self, event):
        self.xml_form['dynamic_form']['mean']['period'] = self.dynamic_form.mean.period_entry.get()

    def dynamic_mean_phase_callback(self, action, sign, unitspage):
        firstlastlist = self.dynamic_form.mean.phase_scrollbar.get()
        if (action == "scroll"):
            if (unitspage == "pages"):
                delta = float(sign) * 1.0 / 4
                setting = firstlastlist[0]
                if ((setting <= 0.0) and (delta < 0.0)):
                    delta = 0.0
                if ((setting >= 1.0) and (delta > 0.0)):
                    delta = 0.0
                setting = setting + delta
                self.dynamic_form.mean.phase_scrollbar.set(setting, setting)
                value = (setting - 0.5) * self.twopi
                self.xml_form['dynamic_form']['mean']['phase'] = value
            else:
                if (unitspage == "units"):
                    delta = float(sign) * 1.0 / 32
                    setting = firstlastlist[0]
                    if ((setting <= 0.0) and (delta < 0.0)):
                        delta = 0.0
                    if ((setting >= 1.0) and (delta > 0.0)):
                        delta = 0.0
                    setting = setting + delta
                    self.dynamic_form.mean.phase_scrollbar.set(setting, setting)
                    value = (setting - 0.5) * self.twopi
                    self.xml_form['dynamic_form']['mean']['phase'] = value
        else:
            if (action == "goto"):
                setting = firstlastlist[0]
                self.dynamic_form.mean.phase_scrollbar.set(setting, setting)
                value = (setting - 0.5) * self.twopi
                self.xml_form['dynamic_form']['mean']['phase'] = value

    def dynamic_range_period_callback(self, event):
        self.xml_form['dynamic_form']['range']['period'] = self.dynamic_form.range.period_entry.get()

    def dynamic_range_phase_callback(self, action, sign, unitspage=None):
        firstlastlist = self.dynamic_form.range.phase_scrollbar.get()
        if (action == "scroll"):
            if (unitspage == "pages"):
                delta = float(sign) * 1.0 / 4
                setting = firstlastlist[0]
                if ((setting <= 0.0) and (delta < 0.0)):
                    delta = 0.0
                if ((setting >= 1.0) and (delta > 0.0)):
                    delta = 0.0
                setting = setting + delta
                self.dynamic_form.range.phase_scrollbar.set(firstlastlist[0] + delta, firstlastlist[0] + delta)
                value = (setting - 0.5) * self.twopi
                self.xml_form['dynamic_form']['range']['phase'] = value
            else:
                if (unitspage == "units"):
                    delta = float(sign) * 1.0 / 32
                    setting = firstlastlist[0]
                    if ((setting <= 0.0) and (delta < 0.0)):
                        delta = 0.0
                    if ((setting >= 1.0) and (delta > 0.0)):
                        delta = 0.0
                    setting = setting + delta
                    self.dynamic_form.range.phase_scrollbar.set(firstlastlist[0] + delta, firstlastlist[0] + delta)
                    value = (setting - 0.5) * self.twopi
                    self.xml_form['dynamic_form']['range']['phase'] = value
        else:
            if (action == "goto"):
                setting = firstlastlist[0]
                self.dynamic_form.range.phase_scrollbar.set(firstlastlist[0])
                value = (setting - 0.5) * self.twopi
                self.xml_form['dynamic_form']['range']['phase'] = value

    def texture_range_period_callback(self, event):
        self.xml_form['texture_form']['period'] = self.texture_form.period_entry.get()

    def texture_range_phase_callback(self, action, sign, unitspage=None):
        firstlastlist = self.texture_form.phase_scrollbar.get()
        if (action == "scroll"):
            if (unitspage == "pages"):
                delta = float(sign) * 1.0 / 4
                setting = firstlastlist[1]
                if ((setting <= 0.0) and (delta < 0.0)):
                    delta = 0.0
                if ((setting >= 1.0) and (delta > 0.0)):
                    delta = 0.0
                setting = setting + delta
                self.texture_form.phase_scrollbar.set(setting, setting)
                value = (setting - 0.5) * self.twopi
                self.xml_form['texture_form']['phase'] = value
            else:
                if (unitspage == "units"):
                    delta = float(sign) * 1.0 / 32
                    setting = firstlastlist[1]
                    if ((setting <= 0.0) and (delta < 0.0)):
                        delta = 0.0
                    if ((setting >= 1.0) and (delta > 0.0)):
                        delta = 0.0
                    setting = setting + delta
                    self.texture_form.phase_scrollbar.set(setting, setting)
                    value = (setting - 0.5) * self.twopi
                    self.xml_form['texture_form']['phase'] = value
        else:
            if (action == "goto"):
                setting = firstlastlist[0]
                self.texture_form.range.phase_scrollbar.set(firstlastlist[0])
                value = (setting - 0.5) * self.twopi
                self.xml_form['texture_form']['phase'] = value

    def install_xml_form(self, xml_form):
        self.xml_form = xml_form

        self.name_entry.delete(0, 1024)
        self.name_entry.insert(0, self.xml_form['name'])
        self.name_entry.update()
        
        self.len_entry.delete(0, 1024)
        self.len_entry.insert(0, self.xml_form['len'])
        self.len_entry.update()

        self.min_note_len_entry.delete(0, 1024)
        self.min_note_len_entry.insert(0, self.xml_form['min_note_len'])
        self.min_note_len_entry.update()

        self.max_note_len_entry.delete(0, 1024)
        self.max_note_len_entry.insert(0, self.xml_form['max_note_len'])
        self.max_note_len_entry.update()

        self.pulse_entry.delete(0, 1024)
        self.pulse_entry.insert(0, self.xml_form['pulse'])
        self.pulse_entry.update()

        self.down_entry.delete(0, 1024)
        self.down_entry.insert(0, self.xml_form['melody_probabilities']['down'])
        self.down_entry.update()

        self.same_entry.delete(0, 1024)
        self.same_entry.insert(0, self.xml_form['melody_probabilities']['same'])
        self.same_entry.update()

        self.up_entry.delete(0, 1024)
        self.up_entry.insert(0, self.xml_form['melody_probabilities']['up'])
        self.up_entry.update()

        self.pitch_form.install_xml_subform(self.xml_form['pitch_form'])
        self.pitch_form.update()
        self.rhythm_form.install_xml_subform(self.xml_form['rhythm_form'])
        self.rhythm_form.update()
        self.dynamic_form.install_xml_subform(self.xml_form['dynamic_form'])
        self.dynamic_form.update()
        self.texture_form.install_xml_sine(self.xml_form['texture_form'])
        self.texture_form.update()
        self.scale_frame.set_keyboard_from_xml(self.xml_form)
        self.scale_frame.update()
        self.update()




#!/usr/bin/env python3
# TextMIDITools Version 1.0.18
# textmidiform.py 1.0
# Copyright © 2022 Thomas E. Janzen
# License GPLv3+: GNU GPL version 3 or later <https://gnu.org/licenses/gpl.html>
# This is free software: you are free to change and redistribute it.
# There is NO WARRANTY, to the extent permitted by law.
#
import tkinter
import tkinter.ttk
from tkinter import *
from tkinter import ttk
from tkinter.ttk import *
from GeneralMIDI import *
import re

class VoiceWindow(tkinter.Toplevel):
    max_voice_number = 23
    general_midi_list = list(GeneralMIDIInstrumentDict)
    def __init__(self, xml_form):
        super().__init__()
        self.frame = ttk.Frame(self, padding="1 1 1 1")
        self.frame.grid()
        self.xml_form = xml_form
        self.create_widgets()
        self.title('Instrumental Voices')
        self.geometry('400x400-50+50')

    def create_widgets(self):
        the_row = 0

        self.number_of_voices_label = tkinter.ttk.Label(self.frame, text="Number of Voices")
        self.number_of_voices_label.grid(row=the_row, column=0, sticky=NSEW, padx=2, pady=2)
        self.number_of_voices_spinbox = tkinter.ttk.Spinbox(self.frame, command=self.number_of_voices_callback, state='readonly', wrap=True)
        self.number_of_voices_spinbox["increment"] = 1
        self.number_of_voices_spinbox["from"]      = 1
        self.number_of_voices_spinbox.grid(row=the_row, column=1, sticky=NSEW, padx=2, pady=2)
        self.number_of_voices_spinbox.bind('<ButtonRelease-1>', self.number_of_voices_callback)
        self.number_of_voices_spinbox['state'] = '!readonly'
        self.number_of_voices_spinbox.set(len(self.xml_form['voices']))
        self.number_of_voices_spinbox["to"] = len(self.xml_form['voices']) - 1
        self.number_of_voices_spinbox['state'] = 'readonly'
        the_row = the_row + 1

        self.voice_number_label = tkinter.ttk.Label(self.frame, text="Voice")
        self.voice_number_label.grid(row=the_row, column=0, sticky=NSEW, padx=2, pady=2)
        self.voice_number_spinbox = tkinter.ttk.Spinbox(self.frame, command=self.voice_number_callback, wrap=True)
        self.voice_number_spinbox["increment"] = 1.0
        self.voice_number_spinbox["from"] = 0.0
        self.voice_number_spinbox["to"] = len(self.xml_form['voices'])
        self.voice_number_spinbox.grid(row=the_row, column=1, sticky=NSEW, padx=2, pady=2)
        self.voice_number_spinbox.bind('<ButtonRelease-1>', self.voice_number_callback)
        vox = 0
        self.voice_number_spinbox.set(vox)
        self.voice_number_spinbox['state'] = 'readonly'
        the_row = the_row + 1

        self.low_pitch_label = tkinter.ttk.Label(self.frame, text="Low Pitch")
        self.low_pitch_label.grid(row=the_row, column=0, sticky=NSEW, padx=2, pady=2)
        # TCL book p 509
        # %d type 1 insert, 0 del, -1 other
        # %i index of string to change or -1
        # %P value of entry if edit is allowed
        # %s val of entry prior to editing
        # %S the text string being inserted or deleted
        # %v type of validation currently set
        # %V type of validation that triggered us (key, focusin, focusout, forced)
        # %W tk name of widget
        validate_command = (self.register(self.validate_pitchname),
            '%d', '%i', '%P', '%s', '%S', '%v', '%V', '%W')
        self.low_pitch_entry = tkinter.ttk.Entry(self.frame, validatecommand=validate_command, validate="focusout")
        self.low_pitch_entry.delete(0, 1024)
        self.low_pitch_entry.insert(0, self.xml_form['voices'][vox]['low_pitch'])
        self.low_pitch_entry.update()
        self.low_pitch_entry.grid(row=the_row, column=1, sticky=NSEW, padx=2, pady=2)
        self.low_pitch_entry.bind('<Key-Return>', self.low_pitch_callback)
        self.low_pitch_entry.bind('<FocusOut>', self.low_pitch_callback)
        the_row = the_row + 1

        self.high_pitch_label = tkinter.ttk.Label(self.frame, text="High Pitch")
        self.high_pitch_label.grid(row=the_row, column=0, sticky=NSEW, padx=2, pady=2)
        self.high_pitch_entry = tkinter.ttk.Entry(self.frame, validatecommand=validate_command, validate="focusout")
        self.high_pitch_entry.delete(0, 1024)
        self.high_pitch_entry.insert(0, self.xml_form['voices'][vox]['high_pitch'])
        self.high_pitch_entry.update()
        self.high_pitch_entry.grid(row=the_row, column=1, sticky=NSEW, padx=2, pady=2)
        self.high_pitch_entry.bind('<Key-Return>', self.high_pitch_callback)
        the_row = the_row + 1

        self.channel_label = tkinter.ttk.Label(self.frame, text="Channel")
        self.channel_label.grid(row=the_row, column=0, sticky=NSEW, padx=2, pady=2)
        self.channel_spinbox = tkinter.ttk.Spinbox(self.frame, command=self.channel_callback, wrap=True)
        self.channel_spinbox["increment"] = 1.0
        self.channel_spinbox["from"] = 1.0
        self.channel_spinbox["to"] = 16.0
        self.channel_spinbox.set(self.xml_form['voices'][vox]['channel'])
        self.channel_spinbox.grid(row=the_row, column=1, sticky=NSEW, padx=2, pady=2)
        self.channel_spinbox.bind('<ButtonRelease-1>', self.channel_callback)
        self.channel_spinbox['state'] = "readonly"
        the_row = the_row + 1

        self.program_label = tkinter.ttk.Label(self.frame, text="Program")
        self.program_label.grid(row=the_row, column=0, sticky=NSEW, padx=2, pady=2)
        self.program_spinbox = tkinter.ttk.Spinbox(self.frame, values=self.general_midi_list, state='readonly', wrap=True)
        self.program_spinbox["increment"] = 1.0
        self.program_spinbox["from"] = 1.0
        self.program_spinbox["to"] = 128.0
        self.program_spinbox.set(self.general_midi_list[self.xml_form['voices'][vox]['program'] - 1])
        self.program_spinbox.grid(row=the_row, column=1, sticky=NSEW, padx=2, pady=2)
        self.program_spinbox.bind('<ButtonRelease-1>', self.program_callback)
        self.program_spinbox['state'] = 'readonly'
        the_row = the_row + 1

        self.walking_label = tkinter.ttk.Label(self.frame, text="Walking")
        self.walking_label.grid(row=the_row, column=0, sticky=NSEW, padx=2, pady=2)
        self.walking = tkinter.IntVar()
        self.walking = self.xml_form['voices'][vox]['walking']
        self.walking_checkbutton = tkinter.ttk.Checkbutton(self.frame, text="walking",
            variable=self.walking, command=self.walking_callback)
        self.walking_checkbutton.grid(row=the_row, column=1, sticky=NSEW, padx=2, pady=2)
        the_row = the_row + 1

        self.pan_scroll_label = tkinter.ttk.Label(self.frame, text="-pi..................0..................pi")
        self.pan_scroll_label.grid(row=the_row, column=1, sticky=NSEW, padx=2, pady=2)
        the_row = the_row + 1

        self.pan_label = tkinter.ttk.Label(self.frame, text="Pan")
        self.pan_label.grid(row=the_row, column=0, sticky=NSEW, padx=2, pady=2)
        self.pan_scrollbar = tkinter.ttk.Scrollbar(self.frame, command=self.pan_callback)
        self.pan_scrollbar["orient"] = "horizontal"
        self.pan_scrollbar.grid(row=the_row, column=1, sticky=NSEW, padx=2, pady=2)
        temppan = self.xml_form['voices'][vox]['pan'] / 128 + 0.5
        self.pan_scrollbar.set(temppan, temppan)
        the_row = the_row + 1

        self.follow_label = tkinter.ttk.Label(self.frame, text="Follow")
        self.follow_label.grid(row=the_row, column=0, sticky=NSEW, padx=2, pady=2)
        self.follow = tkinter.IntVar()
        self.follow.set(self.xml_form['voices'][vox]['follower']['follow'])
        self.follow_checkbutton = tkinter.ttk.Checkbutton(self.frame, text="follow",
            variable=self.follow, command=self.follow_callback)
        self.follow_checkbutton.grid(row=the_row, column=1, sticky=NSEW, padx=2, pady=2)
        the_row = the_row + 1

        self.leader_label = tkinter.ttk.Label(self.frame, text="Leader")
        self.leader_label.grid(row=the_row, column=0, sticky=NSEW, padx=2, pady=2)
        self.leader_spinbox = tkinter.ttk.Spinbox(self.frame, command=self.leader_callback, wrap=True)
        self.leader_spinbox["increment"] = 1.0
        self.leader_spinbox["from"] = 0.0
        self.leader_spinbox["to"] = self.number_of_voices_spinbox.get()
        if (self.xml_form['voices'][vox]['follower']['follow'] == 1):
            self.leader_spinbox["state"] = "enabled"
        else:
            self.leader_spinbox["state"] = "disabled"
        self.leader_spinbox.grid(row=the_row, column=1, sticky=NSEW, padx=2, pady=2)
        self.leader_spinbox.bind('<Key-Return>', self.leader_callback)
        self.leader_spinbox.bind('<FocusOut>', self.leader_callback)
        self.leader_spinbox.bind('<ButtonRelease-1>', self.leader_callback)
        self.leader_spinbox.set(self.xml_form['voices'][vox]['follower']['leader'])
        the_row = the_row + 1

        self.scaler_interval_type_label = tkinter.ttk.Label(self.frame, text="Interval Type")
        self.scaler_interval_type_label.grid(row=the_row, column=0, sticky=NSEW, padx=2, pady=2)
        self.frame.interval_type = tkinter.StringVar()
        self.frame.interval_type.set(self.xml_form['voices'][vox]['follower']['interval_type'])
        self.scaler_interval_type_radiobutton = tkinter.ttk.Radiobutton(self.frame,
            text='Scaler', variable=self.frame.interval_type, value="1",
            command=self.interval_type_callback)
        self.scaler_interval_type_radiobutton.grid(row=the_row, column=1, sticky=NSEW, padx=2, pady=2)

        self.chromatic_interval_type_radiobutton = tkinter.ttk.Radiobutton(self.frame,
            text='Chromatic', variable=self.frame.interval_type, value="2",
            command=self.interval_type_callback)
        if (self.xml_form['voices'][vox]['follower']['follow'] == '1'):
            self.scaler_interval_type_radiobutton["state"] = "enabled"
            self.chromatic_interval_type_radiobutton["state"] = "enabled"
        else:
            self.scaler_interval_type_radiobutton["state"] = "disabled"
            self.chromatic_interval_type_radiobutton["state"] = "disabled"
        self.chromatic_interval_type_radiobutton["state"] = "disabled"
        self.chromatic_interval_type_radiobutton.grid(row=the_row, column=2, sticky=NSEW, padx=2, pady=2)
        the_row = the_row + 1

        self.follow_interval_label = tkinter.ttk.Label(self.frame, text="Interval")
        self.follow_interval_label.grid(row=the_row, column=0, sticky=NSEW, padx=2, pady=2)
        self.follow_interval_spinbox = tkinter.ttk.Spinbox(self.frame, command=self.follow_interval_callback, wrap=True)
        self.follow_interval_spinbox["increment"] = 1.0
        self.follow_interval_spinbox["from"] = -24.0
        self.follow_interval_spinbox["to"] = 24.0
        self.follow_interval_spinbox["state"] = "disabled"
        self.follow_interval_spinbox.grid(row=the_row, column=1, sticky=NSEW, padx=2, pady=2)
        self.follow_interval_spinbox.bind('<ButtonRelease-1>', self.follow_interval_callback)
        self.follow_interval_spinbox.set(self.xml_form['voices'][vox]['follower']['interval'])
        the_row = the_row + 1

    def low_pitch_callback(self, event):
        vox = int(self.voice_number_spinbox.get())
        self.xml_form['voices'][vox]['low_pitch'] = self.low_pitch_entry.get()

    def high_pitch_callback(self, event):
        vox = int(self.voice_number_spinbox.get())
        self.xml_form['voices'][vox]['high_pitch'] = self.high_pitch_entry.get()

    def follow_interval_callback(self, event=None):
        vox = int(self.voice_number_spinbox.get())
        self.xml_form['voices'][vox]['follower']['interval'] = self.follow_interval_spinbox.get()

    def interval_type_callback(self):
        vox = int(self.voice_number_spinbox.get())
        self.xml_form['voices'][vox]['follower']['interval_type'] = self.frame.interval_type.get()

    def number_of_voices_callback(self, event=None):
        vox = int(self.voice_number_spinbox.get())
        number_of_voices = int(self.number_of_voices_spinbox.get())
        if (vox >= number_of_voices):
            vox = (number_of_voices - 1)

        if (len(self.xml_form['voices']) > number_of_voices):
            self.xml_form['voices'] = self.xml_form['voices'][0:(number_of_voices - 1)]
        if (len(self.xml_form['voices']) < number_of_voices):
            for v in range(len(self.xml_form['voices']), number_of_voices):
                vd = {}
                vd['low_pitch'] = "A0"
                vd['high_pitch'] = "C8"
                vd['channel'] = 1
                vd['walking'] = 0
                vd['program'] = 1
                vd['pan'] = 0
                fd = {}
                fd['follow'] = 0
                fd['leader'] = 2147483647
                fd['interval_type'] = 0
                fd['interval'] = 0
                vd['follower'] = fd
                self.xml_form['voices'].append(vd)
        self.voice_number_spinbox["to"] = len(self.xml_form['voices']) - 1

    def voice_number_callback(self, event=None):
        self.update_idletasks()
        vox = int(self.voice_number_spinbox.get())
        ch = self.xml_form['voices'][vox]['channel']
        self.channel_spinbox.set(ch)
        self.channel_spinbox.update()
        instrument = self.general_midi_list[self.xml_form['voices'][vox]['program'] - 1]
        self.program_spinbox['state'] = '!readonly'
        self.program_spinbox.set(instrument)
        self.program_spinbox.update()
        self.program_spinbox['state'] = 'readonly'

        self.low_pitch_entry.delete(0, 1024)
        self.low_pitch_entry.insert(0, self.xml_form['voices'][vox]['low_pitch'])
        self.low_pitch_entry.update()

        self.high_pitch_entry.delete(0, 1024)
        self.high_pitch_entry.insert(0, self.xml_form['voices'][vox]['high_pitch'])
        self.high_pitch_entry.update()

        self.walking = self.xml_form['voices'][vox]['walking']
        if (self.walking_checkbutton.instate(['selected']) and not self.walking):
            self.walking_checkbutton.invoke()
        if (not self.walking_checkbutton.instate(['selected']) and self.walking):
            self.walking_checkbutton.invoke()

        the_pan = float(self.xml_form['voices'][vox]['pan']) / 128.0 + 0.5
        self.pan_scrollbar.set(the_pan, the_pan)
        self.pan_scrollbar.update()

        self.follow = self.xml_form['voices'][vox]['follower']['follow']
        if (self.follow_checkbutton.instate(['selected']) and not self.follow):
            self.follow_checkbutton.invoke()
        if (not self.follow_checkbutton.instate(['selected']) and self.follow):
            self.follow_checkbutton.invoke()

        self.leader_spinbox.set(self.xml_form['voices'][vox]['follower']['leader'])
        self.leader_spinbox.update()

        self.frame.interval_type.set(self.xml_form['voices'][vox]['follower']['interval_type'])
        if (self.xml_form['voices'][vox]['follower']['follow'] == 1):
            self.leader_spinbox["state"] = "enabled"
            self.scaler_interval_type_radiobutton["state"] = "enabled"
            self.chromatic_interval_type_radiobutton["state"] = "enabled"
            self.follow_interval_spinbox.set(self.xml_form['voices'][vox]['follower']['interval'])
            self.frame.interval_type.set(self.xml_form['voices'][vox]['follower']['interval_type'])
            self.scaler_interval_type_radiobutton.update()
            self.chromatic_interval_type_radiobutton.update()
            self.follow_interval_spinbox.update()
        else:
            self.leader_spinbox["state"] = "disabled"
            self.scaler_interval_type_radiobutton["state"] = "disabled"
            self.chromatic_interval_type_radiobutton["state"] = "disabled"
            self.leader_spinbox.set('2147483647')

    def leader_callback(self, event=None):
        self.xml_form['voices'][int(self.voice_number_spinbox.get())]['follower']['leader'] = self.leader_spinbox.get()

    def channel_callback(self, event=None):
        ch = self.channel_spinbox.get()
        self.xml_form['voices'][int(self.voice_number_spinbox.get())]['channel'] = ch

    def program_callback(self, event):
        prog = self.program_spinbox.get()
        vox = int(self.voice_number_spinbox.get())
        self.xml_form['voices'][int(vox)]['program'] = GeneralMIDIInstrumentDict[prog][0]
        # low pitch
        self.low_pitch_entry.delete(0, 1024)
        self.low_pitch_entry.insert(0, GeneralMIDIInstrumentDict[prog][1])
        self.low_pitch_entry.update()
        self.xml_form['voices'][vox]['low_pitch'] = self.low_pitch_entry.get()
        # high pitch

        self.high_pitch_entry.delete(0, 1024)
        self.high_pitch_entry.insert(0, GeneralMIDIInstrumentDict[prog][2])
        self.high_pitch_entry.update()
        self.xml_form['voices'][vox]['high_pitch'] = self.high_pitch_entry.get()

    def walking_callback(self):
        vox = int(self.voice_number_spinbox.get())
        if (self.walking_checkbutton.instate(['selected'])):
            self.xml_form['voices'][vox]['walking'] = 1
        else:
            self.xml_form['voices'][vox]['walking'] = 0

    def install_xml_form(self, xml_form):
        vox = 0
        ch = self.xml_form['voices'][vox]['channel']
        self.xml_form = xml_form
        self.number_of_voices_spinbox['state'] = '!readonly'
        self.number_of_voices_spinbox.set(len(self.xml_form['voices']))
        self.number_of_voices_spinbox['state'] = 'readonly'
        self.voice_number_spinbox['state'] = '!readonly'
        self.voice_number_spinbox["to"] = len(self.xml_form['voices']) - 1
        self.voice_number_spinbox.set(vox)
        self.voice_number_spinbox['state'] = 'readonly'

        self.low_pitch_entry.delete(0, 1024)
        self.low_pitch_entry.insert(0, self.xml_form['voices'][vox]['low_pitch'])
        self.low_pitch_entry.update()

        self.high_pitch_entry.delete(0, 1024)
        self.high_pitch_entry.insert(0, self.xml_form['voices'][vox]['high_pitch'])
        self.high_pitch_entry.update()

        self.channel_spinbox.set(ch)
        self.channel_spinbox.update()
        self.program_spinbox.set(
            self.general_midi_list[self.xml_form['voices'][vox]['program'] - 1])
        self.program_spinbox.update()

        self.walking = self.xml_form['voices'][vox]['walking']
        if (self.walking_checkbutton.instate(['selected']) and not self.walking):
            self.walking_checkbutton.invoke()
        if (not self.walking_checkbutton.instate(['selected']) and self.walking):
            self.walking_checkbutton.invoke()

        the_pan = (float(self.xml_form['voices'][0]['pan']) / 128) + 0.5
        self.pan_scrollbar.set(the_pan, the_pan)
        self.pan_scrollbar.update()

        self.follow = self.xml_form['voices'][vox]['follower']['follow']
        if (self.follow_checkbutton.instate(['selected']) and not self.follow):
            self.follow_checkbutton.invoke()
        if (not self.follow_checkbutton.instate(['selected']) and self.follow):
            self.follow_checkbutton.invoke()

        self.leader_spinbox.set(self.xml_form['voices'][vox]['follower']['leader'])
        self.leader_spinbox.update()
        if (self.xml_form['voices'][vox]['follower']['interval_type'] == 0):
            self.scaler_interval_type_radiobutton["state"] = "disabled"
            self.scaler_interval_type_radiobutton["state"] = "disabled"
        if (self.xml_form['voices'][vox]['follower']['interval_type'] == 1):
            self.scaler_interval_type_radiobutton["state"] = "enabled"
            self.scaler_interval_type_radiobutton["state"] = "disabled"
        if (self.xml_form['voices'][vox]['follower']['interval_type'] == 2):
            self.scaler_interval_type_radiobutton["state"] = "disabled"
            self.scaler_interval_type_radiobutton["state"] = "enabled"
        self.scaler_interval_type_radiobutton.update()
        self.follow_interval_spinbox.set(self.xml_form['voices'][vox]['follower']['interval'])
        self.follow_interval_spinbox.update()
        self.frame.update()

    def follow_callback(self):
        vox = int(self.voice_number_spinbox.get())
        if (self.follow_checkbutton.instate(['selected'])):
            self.xml_form['voices'][vox]['follower']['follow'] = 1
            self.walking_checkbutton["state"] = "disabled"
            self.leader_spinbox["state"] = "normal"
            self.scaler_interval_type_radiobutton["state"] = "normal"
            self.chromatic_interval_type_radiobutton["state"] = "normal"
            self.follow_interval_spinbox["state"] = "normal"
        else:
            self.xml_form['voices'][vox]['follower']['follow'] = 0
            self.walking_checkbutton["state"] = "normal"
            self.leader_spinbox["state"] = "disabled"
            self.scaler_interval_type_radiobutton["state"] = "disabled"
            self.chromatic_interval_type_radiobutton["state"] = "disabled"
            self.follow_interval_spinbox["state"] = "disabled"

    def pan_callback(self, action, sign, unitspage):
        vox = int(self.voice_number_spinbox.get())
        firstlastlist = self.pan_scrollbar.get()
        setting = 0
        if (action == "scroll"):
            if (unitspage == "pages"):
                delta = float(sign) * 1.0 / 4
                setting = firstlastlist[1]
                if ((setting <= 0.0) and (delta < 0.0)):
                    delta = 0.0
                if ((setting >= 1.0) and (delta > 0.0)):
                    delta = 0.0
                setting = setting + delta
                self.pan_scrollbar.set(setting, setting)
                value = (setting - 0.5) * 128
                if (value >= 64):
                    value = 63
                self.xml_form['voices'][vox]['pan'] = value
            else:
                if (unitspage == "units"):
                    delta = float(sign) * 1.0 / 32
                    setting = firstlastlist[1]
                    if ((setting <= 0.0) and (delta < 0.0)):
                        delta = 0.0
                    if ((setting >= 1.0) and (delta > 0.0)):
                        delta = 0.0
                    setting = setting + delta
                    self.pan_scrollbar.set(setting, setting)
                    value = (setting - 0.5) * 128
                    if (value >= 64):
                        value = 63
                    self.xml_form['voices'][vox]['pan'] = value
        else:
            if (action == "goto"):
                setting = firstlastlist[0]
                self.pan_scrollbar.set(firstlastlist[0])
                value = (setting - 0.5) * 128
                if (value >= 64):
                    value = 63
                self.xml_form['voices'][vox]['pan'] = value

    def validate_pitchname(self, d, i, P, s, S, v, V, W):
        # 1-insert, 0=del, -1 if forced
        pat = re.compile('[a-gA-G]([#bx]|bb)-?\d')
        ma = pat.match(str(P))
        ret = False
        if (ma):
            ret = True
        else:
            ret = False
        return ret




#!/usr/bin/env python3
# TextMIDITools Version 1.0.18
# Copyright © 2022 Thomas E. Janzen
# License GPLv3+: GNU GPL version 3 or later <https://gnu.org/licenses/gpl.html>
# This is free software: you are free to change and redistribute it.
# There is NO WARRANTY, to the extent permitted by law.
#

import re
import math
import sys
import tkinter
import tkinter.ttk
from tkinter import *
from tkinter import ttk
from tkinter.ttk import *
import tkinter.constants, tkinter.filedialog

class KeyboardButtons(tkinter.Frame):
    def __init__(self, parent, keyboard):
        super().__init__(parent)
        self.frame = ttk.Frame(self, padding="1 1 1 1")
        self.frame.grid(sticky=NSEW, row=1, column=0)
        self.draw_widgets()
        self.keyboard = keyboard

    def draw_widgets(self):
        self.clear_button = tkinter.ttk.Button(self.frame, text='Clear',
            command=self.clear_keyboard_callback)
        self.clear_button.grid(sticky=NSEW, row=1, column=0)

        self.all_button = tkinter.ttk.Button(self.frame, text='All', command=self.all_keyboard_callback)
        self.all_button.grid(sticky=NSEW, row=1, column=1)

        self.complement_button = tkinter.ttk.Button(self.frame, text='Complement',
            command=self.complement_keyboard_callback)
        self.complement_button.grid(sticky=NSEW, row=1, column=2)

        self.octave_repeat_button = tkinter.ttk.Button(self.frame, text='Octave Repeat',
            command=self.octave_repeat_keyboard_callback)
        self.octave_repeat_button.grid(sticky=NSEW, row=1, column=3)

    def clear_keyboard_callback(self):
        self.keyboard.draw_keyboard()
        for k in range(0, len(self.keyboard.midi_key_selects)):
            self.keyboard.midi_key_selects[k] = False
        self.keyboard.set_keys_from_selects()

    def all_keyboard_callback(self):
        for k in range(0, len(self.keyboard.midi_key_selects)):
            self.keyboard.midi_key_selects[k] = True
        self.keyboard.set_keys_from_selects()

    def complement_keyboard_callback(self):
        for k in range(0, len(self.keyboard.midi_key_selects)):
            self.keyboard.midi_key_selects[k] = not self.keyboard.midi_key_selects[k]
        self.keyboard.set_keys_from_selects()

    def octave_repeat_keyboard_callback(self):
        for k in range(0, len(self.keyboard.midi_key_selects)):
            if (self.keyboard.midi_key_selects[k]):
                for s in range(k % 12, 128, 12):
                    self.keyboard.midi_key_selects[s] = True
        self.keyboard.set_keys_from_selects()

class OneOctave():
    full_octaves = 10
    remainder_keys = 8
    left = 0
    top = 0
    white_key_width = 100
    middle = 403
    bottom = 613
    b_right = white_key_width * 7
    black_key_width = 60
    c_sharp_left = 60
    c_sharp_right = c_sharp_left + black_key_width
    d_left = white_key_width
    d_sharp_left = 180
    d_sharp_right = d_sharp_left + black_key_width
    e_left = 2 * white_key_width
    f_left = 3 * white_key_width
    f_sharp_left  = 356
    f_sharp_right = f_sharp_left + black_key_width
    g_left = 4 * white_key_width
    g_sharp_left = 470
    g_sharp_right = g_sharp_left + black_key_width
    a_left = 5 * white_key_width
    a_sharp_left = 583
    a_sharp_right = a_sharp_left + black_key_width
    b_left = 6 * white_key_width
    ref_keys = [
            ['C', 'white', [[left, top], [left, bottom],
                            [d_left, bottom], [white_key_width, middle],
                            [c_sharp_left, middle], [c_sharp_left, top]]],
            ['C#', 'black', [[c_sharp_left, top], [c_sharp_left, middle],
                             [c_sharp_right, middle], [c_sharp_right, top]]],
            ['D', 'white', [[c_sharp_right, top],[c_sharp_right, middle],
                            [d_left, middle], [d_left, bottom],
                            [e_left, bottom], [e_left, middle],
                            [d_sharp_left, middle], [d_sharp_left, top]]],
            ['D#', 'black', [[d_sharp_left, top], [d_sharp_left, middle],
                             [d_sharp_right, middle], [d_sharp_right, top]]],
            ['E', 'white', [[d_sharp_right, top], [d_sharp_right, middle],
                            [e_left, middle], [e_left, bottom],
                            [f_left, bottom], [f_left, top]]],
            ['F', 'white', [[f_left, top], [f_left, bottom],
                            [g_left, bottom], [g_left, middle],
                            [f_sharp_left, middle], [f_sharp_left, top]]],
            ['F#', 'black', [[f_sharp_left, top], [f_sharp_left, middle],
                             [f_sharp_right, middle], [f_sharp_right, top]]],
            ['G', 'white', [[f_sharp_right, top],[f_sharp_right, middle],
                            [g_left, middle], [g_left, bottom],
                            [a_left, bottom], [a_left, middle],
                            [g_sharp_left, middle], [g_sharp_left, top]]],
            ['G#', 'black', [[g_sharp_left, top], [g_sharp_left, middle],
                             [g_sharp_right, middle], [g_sharp_right, top]]],
            ['A', 'white', [[g_sharp_right, top], [g_sharp_right, middle],
                            [a_left, middle], [a_left, bottom],
                            [b_left, bottom], [b_left, middle],
                            [a_sharp_left, middle], [a_sharp_left, top]]],
            ['A#', 'black', [[a_sharp_left, top], [a_sharp_left, middle],
                             [a_sharp_right, middle], [a_sharp_right, top]]],
            ['B', 'white', [[a_sharp_right, top], [a_sharp_right, middle],
                            [b_left, middle], [b_left, bottom],
                            [b_right, bottom], [b_right, top]]]
        ]

    def __init__(self):
        self.keys = []
        for k in range(0, len(self.ref_keys)):
            temp_key = []
            temp_key.append(self.ref_keys[k][0])
            temp_key.append(self.ref_keys[k][1])
            temp_poly = []
            for p in range(0, len(self.ref_keys[k][2])):
                temp_point = []
                for c in range(0, len(self.ref_keys[k][2][p])):
                    temp_point.append(self.ref_keys[k][2][p][c])
                temp_poly.append(temp_point)
            temp_key.append(temp_poly)
            self.keys.append(temp_key)

    def scale(self, scaler):
        for k in range(0, len(self.ref_keys)):
            for p in range(0, len(self.ref_keys[k][2])):
                for c in range(0, len(self.ref_keys[k][2][p])):
                    self.keys[k][2][p][c] = int(scaler * self.ref_keys[k][2][p][c])

    def translate_one_octave(self):
        self.translate_x(self.keys[11][2][5][0] - self.keys[0][2][0][0])

    def translate_x(self, delta_x):
        for k in range(0, len(self.keys)):
            for p in range(0, len(self.keys[k][2])):
                self.keys[k][2][p][0] = int(delta_x + self.keys[k][2][p][0])

class KeyboardFrame(tkinter.Frame):
    number_of_white_keys = 75
    full_keyboard = []
    midi_key_selects = []

    def __init__(self, parent):
        super().__init__(parent)
        self.frame = ttk.Frame(self, padding="1 1 1 1")
        for k in range(0, 128):
            self.midi_key_selects.append(False)

        self.frame.grid(sticky=NSEW, row=0, column=0)
        self.one_octave = OneOctave()
        self.one_octave.scale(0.20)
        self.canvas_width = self.one_octave.keys[0][2][3][0] * self.number_of_white_keys
        self.canvas_height = self.one_octave.keys[0][2][1][1]
        self.canvas = Canvas(self.frame, bg='white', height=self.canvas_height, width=self.canvas_width, background='#88AAFF')
        self.canvas.grid(row=0, column=0)
        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)
        self.canvas.bind('<Button-1>', self.on_click)
        self.canvas.bind("<Configure>", self.configure_callback)

    def on_click(self, event):
        the_id = self.canvas.find('closest',  event.x, event.y)
        midi_key = (the_id[0] - 1) % 128
        self.midi_key_selects[midi_key] = not self.midi_key_selects[midi_key]
        self.set_keys_from_selects()

    def configure_callback(self, event=None):
        self.win_width = event.width
        self.canvas_width = event.width
        self.win_height = event.height
        self.canvas_height = event.height + 100
        self.draw_keyboard()

    def draw_keyboard(self):
        self.full_keyboard = []

        for octave in range(0, self.one_octave.full_octaves):
            for k in range(0, len(self.one_octave.keys)):
                temp_key = []
                temp_key.append(self.one_octave.keys[k][0] + str(octave - 1))
                temp_key.append(self.one_octave.keys[k][1])
                temp_poly = []
                for p in range(0, len(self.one_octave.keys[k][2])):
                    temp_point = []
                    for c in range(0, len(self.one_octave.keys[k][2][p])):
                        temp_point.append(self.one_octave.keys[k][2][p][c])
                    temp_poly.append(temp_point)
                temp_key.append(temp_poly)
                self.full_keyboard.append(temp_key)
            self.one_octave.translate_one_octave()

        # put in the C to G in the top octave
        octave = 10
        for k in range(0, 8):
            temp_key = []
            temp_key.append(self.one_octave.keys[k][0] + str(octave - 1))
            temp_key.append(self.one_octave.keys[k][1])
            temp_poly = []
            for p in range(0, len(self.one_octave.keys[k][2])):
                temp_point = []
                for c in range(0, len(self.one_octave.keys[k][2][p])):
                    temp_point.append(self.one_octave.keys[k][2][p][c])
                temp_poly.append(temp_point)
            temp_key.append(temp_poly)
            self.full_keyboard.append(temp_key)

        self.canvas.delete(all)
        key_ids = []
        for k in range(0, len(self.full_keyboard)):
            the_width = "1.0"
            if (60 == k):
                the_width = "4.0"
            key_ids.append(self.canvas.create_polygon(self.full_keyboard[k][2], fill=self.full_keyboard[k][1],
                activefill='yellow', disabledfill='gray', width=the_width, outline='black', tags=(self.full_keyboard[k][0])))
        self.set_keys_from_selects()

    def set_keys_from_selects(self):
        if (len(self.full_keyboard) > 0):
            for k in range(0, len(self.midi_key_selects)):
                if (self.midi_key_selects[k]):
                    if (self.full_keyboard[k][1] == 'white'):
                        self.canvas.itemconfigure(tagOrId=self.full_keyboard[k][0], fill='blue', stipple='gray50')
                    else:
                        self.canvas.itemconfigure(tagOrId=self.full_keyboard[k][0], fill='red', stipple='gray50')
                else:
                    self.canvas.itemconfigure(tagOrId=self.full_keyboard[k][0], fill=self.full_keyboard[k][1], stipple='')

    def set_selects(self, selects):
        if ((len(selects) > 0) and (len(selects) == len(self.midi_key_selects))):
            for k in range(0, len(self.midi_key_selects)):
                self.midi_key_selects[k] = selects[k]
            self.set_keys_from_selects()

    def get_key_ints(self):
        key_ints = []
        for flag_index in range(0, len(self.midi_key_selects)):
            if (self.midi_key_selects[flag_index]):
                key_ints.append(flag_index)
        return key_ints

class KeyboardWindow(tkinter.Toplevel):
    def __init__(self):
        super().__init__()
        self.frame = ttk.Frame(self, padding="1 1 1 1")
        self.frame.grid(sticky=NSEW, row=0, column=0)
        self.create_widgets()
        self.title('Keyboard')

    def create_widgets(self):
        self.keyboard_frame = KeyboardFrame(self.frame)
        self.keyboard_frame.grid(sticky=NSEW, row=0, column=0)
        self.keyboard_frame.rowconfigure(index=0, weight=1)
        self.keyboard_frame.columnconfigure(index=0, weight=1)

        self.keyboard_buttons_frame = KeyboardButtons(self.frame, self.keyboard_frame)
        self.keyboard_buttons_frame.grid(sticky=NSEW, row=1, column=0)
        self.keyboard_buttons_frame.rowconfigure(index=1, weight=1)
        self.keyboard_buttons_frame.columnconfigure(index=0, weight=1)

if __name__ == '__main__':
    root = tkinter.Tk()
    keyboard_window = KeyboardWindow()

    root.title('keyboard')

    keyboard_window.mainloop()
